# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""form tests."""
import datetime
import unittest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from libs import form


@test_utils.with_cloud_emulators('datastore')
class GenerateCrsfTokenTest(unittest.TestCase):
  """Test generate_csrf_token."""

  _NOW = datetime.datetime.utcfromtimestamp(10000)
  _LATER = datetime.datetime.utcfromtimestamp(10001)
  _BEFORE = datetime.datetime.utcfromtimestamp(9999)

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.utcnow',
        'libs.helpers.get_user_email',
        'os.urandom',
    ])
    self.mock.get_user_email.return_value = 'test@test.com'

  def make_token(self, value, time):
    token = data_types.CSRFToken()
    token.value = value
    token.expiration_time = time
    token.user_email = self.mock.get_user_email.return_value
    return token

  def test_valid_token(self):
    """Ensure it returns the valid token when the valid token is there."""
    self.mock.utcnow.return_value = self._NOW
    self.make_token(value='token', time=self._LATER).put()
    self.assertEqual(form.generate_csrf_token(), 'token')

  def test_valid_token_html(self):
    """Ensure it returns the valid token in HTML when the valid token is
    there."""
    self.mock.utcnow.return_value = self._NOW
    self.make_token(value='token', time=self._LATER).put()
    self.assertEqual(
        form.generate_csrf_token(html=True),
        '<input type="hidden" name="csrf_token" value="token" />')

  def test_invalid_token(self):
    """Ensure it returns a new token when the valid token expires"""
    self.mock.utcnow.return_value = self._NOW
    self.mock.urandom.return_value = b'a'
    self.make_token(value='token', time=self._BEFORE).put()
    self.assertEqual(form.generate_csrf_token(), 'YQ==')

    tokens = list(
        data_types.CSRFToken.query(
            data_types.CSRFToken.user_email == 'test@test.com'))
    self.assertEqual(len(tokens), 1)
    self.assertGreater(tokens[0].expiration_time, self._NOW)
    self.assertEqual(tokens[0].value, 'YQ==')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Environment functions."""

import ast
import functools
import os
import re
import socket
import subprocess
import sys

import six
import yaml

# Tools supporting customization of options via ADDITIONAL_{TOOL_NAME}_OPTIONS.
# FIXME: Support ADDITIONAL_UBSAN_OPTIONS and ADDITIONAL_LSAN_OPTIONS in an
# ASAN instrumented build.
SUPPORTED_MEMORY_TOOLS_FOR_OPTIONS = [
    'HWASAN', 'ASAN', 'KASAN', 'CFI', 'MSAN', 'TSAN', 'UBSAN'
]

SANITIZER_NAME_MAP = {
    'ASAN': 'address',
    'CFI': 'cfi',
    'MSAN': 'memory',
    'TSAN': 'thread',
    'UBSAN': 'undefined',
}

COMMON_SANITIZER_OPTIONS = {
    'handle_abort': 1,
    'handle_segv': 1,
    'handle_sigbus': 1,
    'handle_sigfpe': 1,
    'handle_sigill': 1,
    'print_summary': 1,
    'use_sigaltstack': 1,
}


def _eval_value(value_string):
  """Returns evaluated value."""
  try:
    return ast.literal_eval(value_string)
  except:
    # String fallback.
    return value_string


def join_memory_tool_options(options):
  """Joins a dict holding memory tool options into a string that can be set in
  the environment."""
  return ':'.join('%s=%s' % (key, str(value))
                  for key, value in sorted(six.iteritems(options)))


def _maybe_convert_to_int(value):
  """Returns the int representation contained by string |value| if it contains
  one. Otherwise returns |value|."""
  try:
    return int(value)
  except ValueError:
    return value


# Matches anything that isn't an unquoted (ie: not between two single or two
# double quotes) colon.
UNQUOTED_COLON_REGEX = re.compile('((?:[^\'":]|\'[^\']*\'|"[^"]*")+)')


def _parse_memory_tool_options(options_str):
  """Parses memory tool options into a dict."""
  parsed = {}

  for item in UNQUOTED_COLON_REGEX.split(options_str):
    # Regex split can give us empty strings at the beginning and the end. Skip
    # these.
    if not item:
      continue
    # Regex split gives us each ':'. Skip these.
    if item == ':':
      continue
    values = item.split('=', 1)
    if len(values) != 2:
      # TODO(mbarbella): Factor this out of environment, and switch to logging
      # an error and continuing. This error should be recoverable.
      raise ValueError('Invalid memory tool option "%s"' % item)

    option_name = values[0]
    option_value = _maybe_convert_to_int(values[1])
    parsed[option_name] = option_value

  return parsed


def _quote_value_if_needed(value):
  """Quote environment value as needed for certain platforms like Windows."""
  result = value
  if ' ' in result or ':' in result:
    result = '"%s"' % result

  return result


def copy():
  """Return a safe copy of the environment."""
  environment_copy = os.environ.copy()
  return environment_copy


def get_asan_options(redzone_size, malloc_context_size, quarantine_size_mb,
                     bot_platform, leaks, disable_ubsan):
  """Generates default ASAN options."""
  asan_options = {}

  # Default options needed for all cases.
  asan_options['alloc_dealloc_mismatch'] = 0
  asan_options['print_scariness'] = 1
  asan_options['strict_memcmp'] = 0

  # Set provided redzone size.
  if redzone_size:
    asan_options['redzone'] = redzone_size

    # This value is used in determining whether to report OOM crashes or not.
    set_value('REDZONE', redzone_size)

  # Set maximum number of stack frames to report.
  if malloc_context_size:
    asan_options['malloc_context_size'] = malloc_context_size

  # Set quarantine size.
  if quarantine_size_mb:
    asan_options['quarantine_size_mb'] = quarantine_size_mb

  # Test for leaks if this is an LSan-enabled job type.
  if get_value('LSAN') and leaks:
    lsan_options = join_memory_tool_options(get_lsan_options())
    set_value('LSAN_OPTIONS', lsan_options)
    asan_options['detect_leaks'] = 1
  else:
    remove_key('LSAN_OPTIONS')
    asan_options['detect_leaks'] = 0

  # FIXME: Support container overflow on Android.
  if is_android(bot_platform):
    asan_options['detect_container_overflow'] = 0

  # Enable stack use-after-return.
  asan_options['detect_stack_use_after_return'] = 1
  asan_options['max_uar_stack_size_log'] = 16

  # Other less important default options for all cases.
  asan_options.update({
      'allocator_may_return_null': 1,
      'allow_user_segv_handler': 0,
      'check_malloc_usable_size': 0,
      'detect_odr_violation': 0,
      'fast_unwind_on_fatal': 1,
      'print_suppressions': 0,
  })

  # Add common sanitizer options.
  asan_options.update(COMMON_SANITIZER_OPTIONS)

  # FIXME: For Windows, rely on online symbolization since llvm-symbolizer.exe
  # in build archive does not work.
  asan_options['symbolize'] = int(bot_platform == 'WINDOWS')

  # For Android, allow user defined segv handler to work.
  if is_android(bot_platform):
    asan_options['allow_user_segv_handler'] = 1

  # Check if UBSAN is enabled as well for this ASAN build.
  # If yes, set UBSAN_OPTIONS and enable suppressions.
  if get_value('UBSAN'):
    if disable_ubsan:
      ubsan_options = get_ubsan_disabled_options()
    else:
      ubsan_options = get_ubsan_options()

    # Remove |symbolize| explicitly to avoid overridding ASan defaults.
    ubsan_options.pop('symbolize', None)

    set_value('UBSAN_OPTIONS', join_memory_tool_options(ubsan_options))

  return asan_options


def get_cpu_arch():
  """Return cpu architecture."""
  if is_android() and not is_android_emulator():
    # FIXME: Handle this import in a cleaner way.
    from clusterfuzz._internal.platforms import android
    return android.settings.get_cpu_arch()

  # FIXME: Add support for desktop architectures as needed.
  return None


def get_current_memory_tool_var():
  """Get the environment variable name for the current job type's sanitizer."""
  memory_tool_name = get_memory_tool_name(get_value('JOB_NAME'))
  if not memory_tool_name:
    return None

  return memory_tool_name + '_OPTIONS'


def get_memory_tool_options(env_var, default_value=None):
  """Get the current memory tool options as a dict. Returns |default_value| if
  |env_var| isn't set. Otherwise returns a dictionary containing the memory tool
  options and their values."""
  env_value = get_value(env_var)
  if env_value is not None:
    return _parse_memory_tool_options(env_value)

  return default_value


def get_instrumented_libraries_paths():
  """Get the instrumented libraries path for the current sanitizer."""
  memory_tool_name = get_memory_tool_name(get_value('JOB_NAME'))
  if not memory_tool_name:
    return None

  if memory_tool_name == 'MSAN':
    if 'no-origins' in get_value('BUILD_URL', ''):
      memory_tool_name += '_NO_ORIGINS'
    else:
      memory_tool_name += '_CHAINED'

  paths = get_value('INSTRUMENTED_LIBRARIES_PATHS_' + memory_tool_name)
  if not paths:
    return None

  return paths.split(':')


def get_default_tool_path(tool_name):
  """Get the default tool for this platform (from scripts/ dir)."""
  if is_android():
    # For android devices, we do symbolization on the host machine, which is
    # linux. So, we use the linux version of llvm-symbolizer.
    platform_override = 'linux'
  else:
    # No override needed, use default.
    platform_override = None

  tool_filename = get_executable_filename(tool_name)
  tool_path = os.path.join(
      get_platform_resources_directory(platform_override), tool_filename)
  return tool_path


def get_environment_settings_as_string():
  """Return environment settings as a string. Includes settings for memory
  debugging tools (e.g. ASAN_OPTIONS for ASAN), application binary revision,
  application command line, etc."""
  environment_string = ''

  # Add Android specific variables.
  if is_android():
    # FIXME: Handle this import in a cleaner way.
    from clusterfuzz._internal.platforms import android

    build_fingerprint = get_value(
        'BUILD_FINGERPRINT') or android.settings.get_build_fingerprint()
    environment_string += '[Environment] Build fingerprint: %s\n' % (
        build_fingerprint)

    security_patch_level = get_value(
        'SECURITY_PATCH_LEVEL') or android.settings.get_security_patch_level()
    environment_string += (
        '[Environment] Patch level: %s\n' % security_patch_level)

    environment_string += (
        '[Environment] Local properties file "%s" with contents:\n%s\n' %
        (android.device.LOCAL_PROP_PATH,
         android.adb.read_data_from_file(android.device.LOCAL_PROP_PATH)))

    command_line = get_value('COMMAND_LINE_PATH')
    if command_line:
      environment_string += (
          '[Environment] Command line file "%s" with contents:\n%s\n' %
          (command_line, android.adb.read_data_from_file(command_line)))

    asan_options = get_value('ASAN_OPTIONS')
    if asan_options:
      # FIXME: Need better documentation for Chrome builds. Chrome builds use
      # asan_device_setup.sh and we send this options file path as an include
      # to extra-options parameter.
      sanitizer_options_file_path = (
          android.sanitizer.get_options_file_path('ASAN'))
      environment_string += (
          '[Environment] ASAN options file "%s" with contents:\n%s\n' %
          (sanitizer_options_file_path, asan_options))

  else:
    # For desktop platforms, add |*_OPTIONS| variables from environment.
    for sanitizer_option in get_sanitizer_options_for_display():
      environment_string += '[Environment] %s\n' % sanitizer_option

  return environment_string


def get_sanitizer_options_for_display():
  """Return a list of sanitizer options with quoted values."""
  result = []
  for tool in SUPPORTED_MEMORY_TOOLS_FOR_OPTIONS:
    options_variable = tool + '_OPTIONS'
    options_value = os.getenv(options_variable)
    if not options_value:
      continue
    result.append('{options_variable}={options_value}'.format(
        options_variable=options_variable, options_value=options_value))

  return result


def get_llvm_symbolizer_path():
  """Get the path of the llvm-symbolizer binary."""
  llvm_symbolizer_path = get_value('LLVM_SYMBOLIZER_PATH')

  if llvm_symbolizer_path and os.path.exists(llvm_symbolizer_path):
    # Make sure that llvm symbolizer binary is executable.
    os.chmod(llvm_symbolizer_path, 0o750)

    return_code = subprocess.call(
        [llvm_symbolizer_path, '--help'],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL)
    if return_code == 0:
      # llvm-symbolize works, return it.
      return llvm_symbolizer_path

  # Either
  # 1. llvm-symbolizer was not found in build archive. OR
  # 2. llvm-symbolizer fails due to dependency issue, clang regression, etc.
  # So, use our own version of llvm-symbolizer.
  llvm_symbolizer_path = get_default_tool_path('llvm-symbolizer')

  # Make sure that we have a default llvm-symbolizer for this platform.
  if not os.path.exists(llvm_symbolizer_path):
    return None

  # Make sure that llvm symbolizer binary is executable.
  os.chmod(llvm_symbolizer_path, 0o750)
  return llvm_symbolizer_path


def get_root_directory():
  """Return root directory."""
  return get_value('ROOT_DIR')


def get_startup_scripts_directory():
  """Return path to startup scripts."""
  return os.path.join(get_value('ROOT_DIR'), 'src', 'python', 'bot', 'startup')


def get_config_directory():
  """Return the path to the configs directory."""
  config_dir = get_value('CONFIG_DIR_OVERRIDE')
  if config_dir:
    return config_dir

  if is_running_on_app_engine():
    # Root is already src/appengine.
    return 'config'

  # Running on bot, give path to config folder inside appengine dir.
  return os.path.join(get_root_directory(), 'src', 'appengine', 'config')


def get_gae_config_directory():
  """Return the path to the google appengine configs directory."""
  return os.path.join(get_config_directory(), 'gae')


def get_gce_config_directory():
  """Return the path to the google compute engine configs directory."""
  return os.path.join(get_config_directory(), 'gce')


def get_resources_directory():
  """Return the path to the resources directory."""
  return os.path.join(get_root_directory(), 'resources')


def get_platform_resources_directory(platform_override=None):
  """Return the path to platform-specific resources directory."""
  plt = platform_override or platform()

  # Android resources share the same android directory.
  if is_android(plt):
    plt = 'ANDROID'

  return os.path.join(get_resources_directory(), 'platform', plt.lower())


def get_suppressions_directory():
  """Return the path to the suppressions directory."""
  return os.path.join(get_config_directory(), 'suppressions')


def get_suppressions_file(sanitizer, suffix='suppressions'):
  """Return the path to sanitizer suppressions file, if exists."""
  sanitizer_suppressions_filename = '{sanitizer}_{suffix}.txt'.format(
      sanitizer=sanitizer, suffix=suffix)
  sanitizer_suppressions_file_path = os.path.join(
      get_suppressions_directory(), sanitizer_suppressions_filename)

  if not os.path.exists(sanitizer_suppressions_file_path):
    return None

  if not os.path.getsize(sanitizer_suppressions_file_path):
    return None

  return sanitizer_suppressions_file_path


def get_lsan_options():
  """Generates default LSAN options."""
  lsan_suppressions_path = get_suppressions_file('lsan')
  lsan_options = {
      'print_suppressions': 0,
  }

  # Add common sanitizer options.
  lsan_options.update(COMMON_SANITIZER_OPTIONS)

  if lsan_suppressions_path:
    lsan_options['suppressions'] = lsan_suppressions_path

  return lsan_options


def get_kasan_options():
  """Generates default KASAN options."""
  kasan_options = {'symbolize': 0}

  # Add common sanitizer options.
  kasan_options.update(COMMON_SANITIZER_OPTIONS)

  return kasan_options


def get_msan_options():
  """Generates default MSAN options."""
  msan_options = {'symbolize': 0}

  # Add common sanitizer options.
  msan_options.update(COMMON_SANITIZER_OPTIONS)

  return msan_options


def get_platform_id():
  """Return a platform id as a lowercase string."""
  bot_platform = platform()
  if is_android_cuttlefish() or is_android_emulator():
    return bot_platform.lower()
  if is_android(bot_platform):
    # FIXME: Handle this import in a cleaner way.
    from clusterfuzz._internal.platforms import android

    platform_id = get_value('PLATFORM_ID', android.settings.get_platform_id())
    return platform_id.lower()

  return bot_platform.lower()


def get_platform_group():
  """Return the platform group (specified via QUEUE_OVERRIDE) if it
  exists, otherwise platform()."""
  platform_group = get_value('QUEUE_OVERRIDE')
  if platform_group:
    return platform_group

  return platform()


def get_memory_tool_name(job_name):
  """Figures out name of memory debugging tool."""
  for tool in SUPPORTED_MEMORY_TOOLS_FOR_OPTIONS:
    if tool_matches(tool, job_name):
      return tool

  # If no tool specified, assume it is ASAN. Also takes care of LSAN job type.
  return 'ASAN'


def get_memory_tool_display_string(job_name):
  """Return memory tool string for a testcase."""
  memory_tool_name = get_memory_tool_name(job_name)
  sanitizer_name = SANITIZER_NAME_MAP.get(memory_tool_name)
  if not sanitizer_name:
    return 'Memory Tool: %s' % memory_tool_name

  return 'Sanitizer: %s (%s)' % (sanitizer_name, memory_tool_name)


def get_executable_filename(executable_name):
  """Return the filename for the given executable."""
  if platform() != 'WINDOWS':
    return executable_name

  extension = '.exe'
  if executable_name.endswith(extension):
    return executable_name

  return executable_name + extension


def get_tsan_options():
  """Generates default TSAN options."""
  tsan_suppressions_path = get_suppressions_file('tsan')

  tsan_options = {
      'atexit_sleep_ms': 200,
      'flush_memory_ms': 2000,
      'history_size': 3,
      'print_suppressions': 0,
      'report_thread_leaks': 0,
      'report_signal_unsafe': 0,
      'stack_trace_format': 'DEFAULT',
      'symbolize': 1,
  }

  # Add common sanitizer options.
  tsan_options.update(COMMON_SANITIZER_OPTIONS)

  if tsan_suppressions_path:
    tsan_options['suppressions'] = tsan_suppressions_path

  return tsan_options


def get_ubsan_options():
  """Generates default UBSAN options."""
  # Note that UBSAN can work together with ASAN as well.
  ubsan_suppressions_path = get_suppressions_file('ubsan')

  ubsan_options = {
      'halt_on_error': 1,
      'print_stacktrace': 1,
      'print_suppressions': 0,

      # We use -fsanitize=unsigned-integer-overflow as an additional coverage
      # signal and do not want those errors to be reported by UBSan as bugs.
      # See https://github.com/google/oss-fuzz/issues/910 for additional info.
      'silence_unsigned_overflow': 1,
      'symbolize': 1,
  }

  # Add common sanitizer options.
  ubsan_options.update(COMMON_SANITIZER_OPTIONS)

  # TODO(crbug.com/877070): Make this code configurable on a per job basis.
  if ubsan_suppressions_path and not is_chromeos_system_job():
    ubsan_options['suppressions'] = ubsan_suppressions_path

  return ubsan_options


def get_ubsan_disabled_options():
  """Generates ubsan options """
  return {
      'halt_on_error': 0,
      'print_stacktrace': 0,
      'print_suppressions': 0,
  }


def get_value(environment_variable, default_value=None):
  """Return an environment variable value."""
  value_string = os.getenv(environment_variable)

  # value_string will be None if the variable is not defined.
  if value_string is None:
    return default_value

  # Exception for ANDROID_SERIAL. Sometimes serial can be just numbers,
  # so we don't want to it eval it.
  if environment_variable == 'ANDROID_SERIAL':
    return value_string

  # Evaluate the value of the environment variable with string fallback.
  return _eval_value(value_string)


def _job_substring_match(search_string, job_name):
  """Return a bool on whether a string exists in a provided job name or
  use from environment if available (case insensitive)."""
  job_name = job_name or get_value('JOB_NAME')
  if not job_name:
    return False

  return search_string in job_name.lower()


def is_afl_job(job_name=None):
  """Return true if the current job uses AFL."""
  # Prefix matching is not sufficient.
  return _job_substring_match('afl', job_name)


def is_chromeos_job(job_name=None):
  """Return True if the current job is for ChromeOS."""
  return _job_substring_match('chromeos', job_name)


def is_lkl_job(job_name=None):
  """Return True if the current job is for ChromeOS."""
  return _job_substring_match('lkl', job_name)


def is_chromeos_system_job(job_name=None):
  """Return True if the current job is for ChromeOS system (i.e. not libFuzzer
  or entire Chrome browser for Chrome on ChromeOS)."""
  return is_chromeos_job(job_name) and get_value('CHROMEOS_SYSTEM')


def is_libfuzzer_job(job_name=None):
  """Return true if the current job uses libFuzzer."""
  # Prefix matching is not sufficient.
  return _job_substring_match('libfuzzer', job_name)


def is_honggfuzz_job(job_name=None):
  """Return true if the current job uses honggfuzz."""
  return _job_substring_match('honggfuzz', job_name)


def is_kernel_fuzzer_job(job_name=None):
  """Return true if the current job uses syzkaller."""
  return _job_substring_match('syzkaller', job_name)


def is_googlefuzztest_job(job_name=None):
  """Return true if the current job uses googlefuzztest."""
  return _job_substring_match('googlefuzztest', job_name)


def is_engine_fuzzer_job(job_name=None):
  """Return true if this is an engine fuzzer."""
  return bool(get_engine_for_job(job_name))


def get_engine_for_job(job_name=None):
  """Get the engine for the given job."""
  # TODO(ochang): Generalize this rather than hardcoding all these engines.
  if is_libfuzzer_job(job_name):
    return 'libFuzzer'
  if is_afl_job(job_name):
    return 'afl'
  if is_honggfuzz_job(job_name):
    return 'honggfuzz'
  if is_kernel_fuzzer_job(job_name):
    return 'syzkaller'
  if is_googlefuzztest_job(job_name):
    return 'googlefuzztest'

  return None


def is_posix():
  """Return true if we are on a posix platform (linux/unix and mac os)."""
  return os.name == 'posix'


def is_trusted_host(ensure_connected=True):
  """Return whether or not the current bot is a trusted host."""
  return get_value('TRUSTED_HOST') and (not ensure_connected or
                                        get_value('WORKER_BOT_NAME'))


def is_untrusted_worker():
  """Return whether or not the current bot is an untrusted worker."""
  return get_value('UNTRUSTED_WORKER')


def is_running_on_app_engine():
  """Return True if we are running on appengine (local or production)."""
  return (os.getenv('GAE_ENV') or is_running_on_app_engine_development() or
          os.getenv('SERVER_SOFTWARE', '').startswith('Google App Engine/'))


def is_running_on_app_engine_development():
  """Return True if running on the local development appengine server."""
  return (os.getenv('GAE_ENV') == 'dev' or
          os.getenv('SERVER_SOFTWARE', '').startswith('Development/'))


def parse_environment_definition(environment_string):
  """Parses a job's environment definition."""
  if not environment_string:
    return {}

  definitions = [environment_string.splitlines()]
  values = {}
  for definition in definitions:
    for line in definition:
      if line.startswith('#') or not line.strip():
        continue

      m = re.match('([^ =]+)[ ]*=[ ]*(.*)', line)
      if m:
        key = m.group(1).strip()
        value = m.group(2).strip()
        values[key] = value

  return values


def platform():
  """Return the operating system type, unless an override is provided."""
  environment_override = get_value('OS_OVERRIDE')
  if environment_override:
    return environment_override.upper()

  if sys.platform.startswith('win'):
    return 'WINDOWS'
  if sys.platform.startswith('linux'):
    return 'LINUX'
  if sys.platform == 'darwin':
    return 'MAC'

  raise ValueError('Unsupported platform "%s".' % sys.platform)


def remove_key(key_name):
  """Remove environment |key| and its associated value."""
  if not key_name:
    return

  if key_name not in os.environ:
    return

  del os.environ[key_name]


# Used by reset_environment to store the initial environment.
_initial_environment = None


def reset_environment():
  """Resets environment variables to their initial state. Saves the initial
    state on first call."""
  global _initial_environment
  if _initial_environment is None:
    _initial_environment = copy()
    # There is nothing to reset if we are initializing for the first time.
  else:
    # Clean current environment.
    os.environ.clear()

    # Add shared variables with values from _initial_environment.
    os.environ.update(_initial_environment)

  if is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import \
        environment as untrusted_env
    untrusted_env.reset_environment()


def set_common_environment_variables():
  """Sets environment variables common for different memory debugging tools."""
  # G_SLICE = always-malloc: make glib use system malloc.
  # NSS_DISABLE_UNLOAD = 1: make nss skip dlclosing dynamically loaded modules,
  # which would result in "obj:*" in backtraces.
  # NSS_DISABLE_ARENA_FREE_LIST = 1: make nss use system malloc.
  set_value('G_SLICE', 'always-malloc')
  set_value('NSS_DISABLE_UNLOAD', 1)
  set_value('NSS_DISABLE_ARENA_FREE_LIST', 1)
  set_value('NACL_DANGEROUS_SKIP_QUALIFICATION_TEST', 1)


def set_memory_tool_options(env_var, options_dict):
  """Set current memory tool options."""
  set_value(env_var, join_memory_tool_options(options_dict))


def set_environment_parameters_from_file(file_path):
  """Set environment variables from a file."""
  if not os.path.exists(file_path):
    return

  with open(file_path, 'r') as f:
    file_data = f.read()

  for line in file_data.splitlines():
    if line.startswith('#') or not line.strip():
      continue

    m = re.match('([^ =]+)[ ]*=[ ]*(.*)', line)
    if m:
      environment_variable = m.group(1)
      environment_variable_value = m.group(2)
      set_value(environment_variable, environment_variable_value)


def update_symbolizer_options(tool_options, symbolize_inline_frames=False):
  """Checks and updates the necessary symbolizer options such as
  `external_symbolizer_path` and `symbolize_inline_frames`."""
  if 'external_symbolizer_path' not in tool_options:
    llvm_symbolizer_path = get_llvm_symbolizer_path()
    if llvm_symbolizer_path:
      tool_options.update({
          'external_symbolizer_path':
              _quote_value_if_needed(llvm_symbolizer_path)
      })
  if 'symbolize_inline_frames' not in tool_options:
    tool_options.update({
        'symbolize_inline_frames': str(symbolize_inline_frames).lower()
    })


def reset_current_memory_tool_options(redzone_size=0,
                                      malloc_context_size=0,
                                      leaks=True,
                                      symbolize_inline_frames=False,
                                      quarantine_size_mb=None,
                                      disable_ubsan=False):
  """Resets environment variables for memory debugging tool to default
  values."""
  # FIXME: Handle these imports in a cleaner way.
  from clusterfuzz._internal.platforms import android

  # Set common environment variable useful for memory debugging tools.
  set_common_environment_variables()

  # Set memory tool name in our environment for easy access.
  job_name = get_value('JOB_NAME')
  tool_name = get_memory_tool_name(job_name)
  set_value('MEMORY_TOOL', tool_name)

  bot_platform = platform()

  # Default options for memory debuggin tool used.
  if tool_name in ['ASAN', 'HWASAN']:
    tool_options = get_asan_options(redzone_size, malloc_context_size,
                                    quarantine_size_mb, bot_platform, leaks,
                                    disable_ubsan)
  elif tool_name == 'KASAN':
    tool_options = get_kasan_options()
  elif tool_name == 'MSAN':
    tool_options = get_msan_options()
  elif tool_name == 'TSAN':
    tool_options = get_tsan_options()
  elif tool_name in ['UBSAN', 'CFI']:
    tool_options = get_ubsan_options()

  # Additional options. These override the defaults.
  additional_tool_options = get_value('ADDITIONAL_%s_OPTIONS' % tool_name)
  if additional_tool_options:
    tool_options.update(_parse_memory_tool_options(additional_tool_options))

  if tool_options.get('symbolize') == 1:
    update_symbolizer_options(
        tool_options, symbolize_inline_frames=symbolize_inline_frames)

  # Join the options.
  joined_tool_options = join_memory_tool_options(tool_options)
  tool_options_variable_name = '%s_OPTIONS' % tool_name
  set_value(tool_options_variable_name, joined_tool_options)

  # CFI handles various signals through the UBSan runtime, so need to set
  # UBSAN_OPTIONS explicitly. See crbug.com/716235#c25
  if tool_name == 'CFI':
    set_value('UBSAN_OPTIONS', joined_tool_options)

  # For Android, we need to set shell property |asan.options|.
  # For engine-based fuzzers, it is not needed as options variable is directly
  # passed to shell.
  if is_android(bot_platform) and not is_engine_fuzzer_job():
    android.sanitizer.set_options(tool_name, joined_tool_options)


def set_default_vars():
  """Set default environment vars and values."""
  env_file_path = os.path.join(get_value('ROOT_DIR'), 'bot', 'env.yaml')
  with open(env_file_path) as file_handle:
    env_file_contents = file_handle.read()

  env_vars_and_values = yaml.safe_load(env_file_contents)
  for variable, value in six.iteritems(env_vars_and_values):
    # We cannot call set_value here.
    os.environ[variable] = str(value)


def set_bot_environment():
  """Set environment for the bots."""
  root_dir = get_value('ROOT_DIR')

  if not root_dir:
    # Error, bail out.
    return False

  # Reset our current working directory. Our's last job might
  # have left us in a non-existent temp directory.
  # Or ROOT_DIR might be deleted and recreated.
  os.chdir(root_dir)

  # Set some default directories. These can be overriden by config files below.
  bot_dir = os.path.join(root_dir, 'bot')
  if is_trusted_host(ensure_connected=False):
    worker_root_dir = os.environ['WORKER_ROOT_DIR']
    os.environ['BUILDS_DIR'] = os.path.join(worker_root_dir, 'bot', 'builds')
  else:
    os.environ['BUILDS_DIR'] = os.path.join(bot_dir, 'builds')

  os.environ['BUILD_URLS_DIR'] = os.path.join(bot_dir, 'build-urls')
  os.environ['LOG_DIR'] = os.path.join(bot_dir, 'logs')
  os.environ['CACHE_DIR'] = os.path.join(bot_dir, 'cache')

  inputs_dir = os.path.join(bot_dir, 'inputs')
  os.environ['INPUT_DIR'] = inputs_dir
  os.environ['CRASH_STACKTRACES_DIR'] = os.path.join(inputs_dir, 'crash-stacks')
  os.environ['FUZZERS_DIR'] = os.path.join(inputs_dir, 'fuzzers')
  os.environ['DATA_BUNDLES_DIR'] = os.path.join(inputs_dir, 'data-bundles')
  os.environ['FUZZ_INPUTS'] = os.path.join(inputs_dir, 'fuzzer-testcases')
  os.environ['FUZZ_INPUTS_MEMORY'] = os.environ['FUZZ_INPUTS']
  os.environ['FUZZ_INPUTS_DISK'] = os.path.join(inputs_dir,
                                                'fuzzer-testcases-disk')
  os.environ['MUTATOR_PLUGINS_DIR'] = os.path.join(inputs_dir,
                                                   'mutator-plugins')
  os.environ['FUZZ_DATA'] = os.path.join(inputs_dir,
                                         'fuzzer-common-data-bundles')
  os.environ['IMAGES_DIR'] = os.path.join(inputs_dir, 'images')
  os.environ['SYMBOLS_DIR'] = os.path.join(inputs_dir, 'symbols')
  os.environ['USER_PROFILE_ROOT_DIR'] = os.path.join(inputs_dir,
                                                     'user-profile-dirs')

  # Set bot name.
  if not get_value('BOT_NAME'):
    # If not defined, default to host name.
    os.environ['BOT_NAME'] = socket.gethostname().lower()

  # Local temp directory (non-tmpfs).
  local_tmp_dir = os.path.join(bot_dir, 'tmp')

  # Set BOT_TMPDIR if not already set.
  if not get_value('BOT_TMPDIR'):
    os.environ['BOT_TMPDIR'] = local_tmp_dir

  # Add common environment variables needed by Bazel test runner.
  # See https://docs.bazel.build/versions/master/test-encyclopedia.html.
  # NOTE: Do not use a tmpfs folder as some fuzz targets don't work.
  os.environ['TEST_TMPDIR'] = local_tmp_dir
  os.environ['TZ'] = 'UTC'

  # Sets the default configuration. Can be overridden by job environment.
  set_default_vars()

  # Set environment variable from local project configuration.
  from clusterfuzz._internal.config import local_config
  local_config.ProjectConfig().set_environment()

  # Success.
  return True


def set_tsan_max_history_size():
  """Sets maximum history size for TSAN tool."""
  tsan_options = get_value('TSAN_OPTIONS')
  if not tsan_options:
    return

  tsan_max_history_size = 7
  for i in range(tsan_max_history_size):
    tsan_options = (
        tsan_options.replace('history_size=%d' % i,
                             'history_size=%d' % tsan_max_history_size))

  set_value('TSAN_OPTIONS', tsan_options)


def set_value(environment_variable, value):
  """Set an environment variable."""
  value_str = str(value)
  environment_variable_str = str(environment_variable)
  value_str = value_str.replace('%ROOT_DIR%', os.getenv('ROOT_DIR', ''))
  os.environ[environment_variable_str] = value_str

  if is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import \
        environment as untrusted_env
    untrusted_env.forward_environment_variable(environment_variable_str,
                                               value_str)


def tool_matches(tool_name, job_name):
  """Return if the memory debugging tool is used in this job."""
  match_prefix = '(.*[^a-zA-Z]|^)%s'
  matches_tool = re.match(match_prefix % tool_name.lower(), job_name.lower())
  return bool(matches_tool)


def appengine_noop(func):
  """Wrap a function into no-op and return None if running on App Engine."""

  @functools.wraps(func)
  def wrapper(*args, **kwargs):
    if is_running_on_app_engine():
      return None

    return func(*args, **kwargs)

  return wrapper


def bot_noop(func):
  """Wrap a function into no-op and return None if running on bot."""

  @functools.wraps(func)
  def wrapper(*args, **kwargs):
    is_bot = not is_running_on_app_engine()
    if is_bot:
      return None

    return func(*args, **kwargs)

  return wrapper


def is_local_development():
  """Return true if running in local development environment (e.g. running
  a bot locally, excludes tests)."""
  return bool(get_value('LOCAL_DEVELOPMENT') and not get_value('PY_UNITTESTS'))


def local_noop(func):
  """Wrap a function into no-op and return None if running in local
  development environment."""

  @functools.wraps(func)
  def wrapper(*args, **kwargs):
    if (is_local_development() or is_running_on_app_engine_development()):
      return None

    return func(*args, **kwargs)

  return wrapper


def is_ephemeral():
  """Return whether or not we are an ephemeral bot."""
  return get_value('EPHEMERAL')


def is_android(plt=None):
  """Return true if we are on android platform."""
  return 'ANDROID' in (plt or platform())


def is_android_cuttlefish(plt=None):
  """Return true if we are on android cuttlefish platform."""
  return 'ANDROID_X86' in (plt or platform())


def is_android_emulator(plt=None):
  """Return true if we are on android emulator platform."""
  return 'ANDROID_EMULATOR' == (plt or platform())


def is_android_kernel(plt=None):
  """Return true if we are on android kernel platform groups."""
  return 'ANDROID_KERNEL' in (plt or get_platform_group())


def is_lib():
  """Whether or not we're in libClusterFuzz."""
  return get_value('LIB_CF')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""update_task tests."""
import os
import sys
import unittest

from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.tasks import update_task
from clusterfuzz._internal.metrics import monitor
from clusterfuzz._internal.metrics import monitoring_metrics
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class GetLocalSourceRevisionTest(fake_filesystem_unittest.TestCase):
  """Test get_local_source_revision."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)
    helpers.patch_environ(self)

    os.environ['ROOT_DIR'] = '/root'
    os.environ['FAIL_RETRIES'] = '1'

  def test_no_revision(self):
    """Test when there's no revision."""
    self.assertIsNone(update_task.get_local_source_revision())

  def test_has_revision(self):
    """Test when there's a revision."""
    os.mkdir('/root')
    self.fs.create_file(
        os.path.join('/root', utils.LOCAL_SOURCE_MANIFEST), contents='revision')
    self.assertEqual('revision', update_task.get_local_source_revision())


class TrackRevisionTest(fake_filesystem_unittest.TestCase):
  """Test _track_revision."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)
    helpers.patch_environ(self)

    os.environ['ROOT_DIR'] = '/root'
    os.environ['FAIL_RETRIES'] = '1'

    monitor.metrics_store().reset_for_testing()

  def test_no_revision(self):
    """Test when there's no revision."""
    update_task.track_revision()
    self.assertEqual(0,
                     monitoring_metrics.BOT_COUNT.get({
                         'revision': 'revision'
                     }))

  def test_has_revision(self):
    """Test when there's a revision."""
    os.mkdir('/root')
    self.fs.create_file(
        os.path.join('/root', utils.LOCAL_SOURCE_MANIFEST), contents='revision')
    update_task.track_revision()
    self.assertEqual(1,
                     monitoring_metrics.BOT_COUNT.get({
                         'revision': 'revision'
                     }))


class GetNewerSourceRevisionTest(unittest.TestCase):
  """Test get_newer_source_revision."""

  def setUp(self):
    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.bot.tasks.update_task.get_remote_source_revision',
        'clusterfuzz._internal.bot.tasks.update_task.get_local_source_revision',
    ])

    os.environ['ROOT_DIR'] = 'root_dir'
    os.environ['BOT_TMPDIR'] = 'bot_tmpdir'
    os.environ['FAIL_RETRIES'] = '1'

    if sys.version_info.major == 3:
      self.manifest_suffix = '.3'
    else:
      self.manifest_suffix = ''

  def test_no_local_revision(self):
    """Test no local revision."""
    self.mock.get_remote_source_revision.return_value = 'remote'
    self.mock.get_local_source_revision.return_value = ''
    self.assertEqual('remote', update_task.get_newer_source_revision())

    self.mock.get_local_source_revision.assert_called_once_with()
    self.mock.get_remote_source_revision.assert_called_once_with(
        'gs://test-deployment-bucket/clusterfuzz-source.manifest' +
        self.manifest_suffix)

  def test_error_on_remote_revision(self):
    """Test error on remote revision."""
    self.mock.get_remote_source_revision.side_effect = Exception('fake')
    self.assertIsNone(update_task.get_newer_source_revision())

    self.assertEqual(0, self.mock.get_local_source_revision.call_count)

  def test_older_revision(self):
    """Test remote revision is older."""
    self.mock.get_remote_source_revision.return_value = '123'
    self.mock.get_local_source_revision.return_value = '456789'
    self.assertIsNone(update_task.get_newer_source_revision())

    self.mock.get_local_source_revision.assert_called_once_with()
    self.mock.get_remote_source_revision.assert_called_once_with(
        'gs://test-deployment-bucket/clusterfuzz-source.manifest' +
        self.manifest_suffix)

  def test_newer_revision(self):
    """Test remote revision is newer."""
    self.mock.get_remote_source_revision.return_value = '456'
    self.mock.get_local_source_revision.return_value = '12345'
    self.assertEqual('456', update_task.get_newer_source_revision())

    self.mock.get_local_source_revision.assert_called_once_with()
    self.mock.get_remote_source_revision.assert_called_once_with(
        'gs://test-deployment-bucket/clusterfuzz-source.manifest' +
        self.manifest_suffix)


class GetUrlsTest(unittest.TestCase):
  """Tests getting source and manifest URLs."""

  def setUp(self):
    helpers.patch_environ(self)
    helpers.patch(self, [
        'platform.system',
    ])
    if sys.version_info.major == 3:
      self.manifest_suffix = '.3'
      self.deployment_suffix = '-3'
    else:
      self.manifest_suffix = ''
      self.deployment_suffix = ''

  def test_get_source_manifest_url(self):
    """Test get_source_manifest_url."""
    self.assertEqual(
        'gs://test-deployment-bucket/clusterfuzz-source.manifest' +
        self.manifest_suffix, update_task.get_source_manifest_url())

  def test_get_source_url_windows(self):
    """Test get_source_url on windows."""
    self.mock.system.return_value = 'Windows'
    self.assertEqual(
        'gs://test-deployment-bucket/windows%s.zip' % self.deployment_suffix,
        update_task.get_source_url())

  def test_get_source_url_macos(self):
    """Test get_source_url on macos."""
    self.mock.system.return_value = 'Darwin'
    self.assertEqual(
        'gs://test-deployment-bucket/macos%s.zip' % self.deployment_suffix,
        update_task.get_source_url())

  def test_get_source_url_linux(self):
    """Test get_source_url on linux."""
    self.mock.system.return_value = 'Linux'
    self.assertEqual(
        'gs://test-deployment-bucket/linux%s.zip' % self.deployment_suffix,
        update_task.get_source_url())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for remote_process."""

import unittest

import mock

from clusterfuzz._internal.bot.untrusted_runner import remote_process
from clusterfuzz._internal.bot.untrusted_runner import remote_process_host
from clusterfuzz._internal.protos import untrusted_runner_pb2
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers


class RemoteProcessTest(unittest.TestCase):
  """RemoteProcess tests."""

  def setUp(self):
    test_helpers.patch(
        self, ['clusterfuzz._internal.system.process_handler.run_process'])

  @mock.patch.object(new_process.ProcessRunner, 'run_and_wait')
  def test_run_and_wait(self, mock_run_and_wait):
    """Test remote_process.run_and_wait()."""
    process_result = new_process.ProcessResult(['command', '123'], 0, b'output',
                                               60.0, False)

    mock_run_and_wait.return_value = process_result

    request = untrusted_runner_pb2.RunAndWaitRequest()
    request.executable_path = '/path'
    request.default_args.extend(['-default_arg'])
    request.additional_args.extend(['-additional_arg'])
    request.timeout = 100.0
    request.terminate_before_kill = True
    request.terminate_wait_time = 10.0
    request.input_data = b'input'
    request.max_stdout_len = 1337
    request.popen_args.shell = True
    request.popen_args.env.update({'VAR': 'VAL'})
    request.popen_args.env_is_set = True
    request.popen_args.cwd = '/'

    response = remote_process.run_and_wait(request, None)
    result = remote_process_host.process_result_from_proto(response.result)

    mock_run_and_wait.assert_called_with(
        additional_args=['-additional_arg'],
        timeout=100.0,
        terminate_before_kill=True,
        terminate_wait_time=10.0,
        input_data=b'input',
        max_stdout_len=1337,
        cwd='/',
        env={'VAR': 'VAL'},
        shell=True)

    self.assertEqual(result.command, process_result.command)
    self.assertEqual(result.return_code, process_result.return_code)
    self.assertEqual(result.output, process_result.output)
    self.assertEqual(result.time_executed, process_result.time_executed)
    self.assertEqual(result.timed_out, process_result.timed_out)

  @mock.patch.object(new_process.ProcessRunner, 'run_and_wait')
  def test_run_and_wait_none_env(self, mock_run_and_wait):
    """Test remote_process.run_and_wait() with a None env."""
    process_result = new_process.ProcessResult(['command', '123'], 0, b'output',
                                               60.0, False)

    mock_run_and_wait.return_value = process_result

    request = untrusted_runner_pb2.RunAndWaitRequest()
    request.executable_path = '/path'
    request.popen_args.env_is_set = False

    remote_process.run_and_wait(request, None)
    mock_run_and_wait.assert_called_with(additional_args=[], env=None)

  @mock.patch.object(new_process.ProcessRunner, 'run_and_wait')
  def test_run_and_wait_empty_env(self, mock_run_and_wait):
    """Test remote_process.run_and_wait() with an empty env."""
    process_result = new_process.ProcessResult(['command', '123'], 0, b'output',
                                               60.0, False)

    mock_run_and_wait.return_value = process_result

    request = untrusted_runner_pb2.RunAndWaitRequest()
    request.executable_path = '/path'
    request.popen_args.env_is_set = True

    remote_process.run_and_wait(request, None)
    mock_run_and_wait.assert_called_with(additional_args=[], env={})

  def test_run_process(self):
    """Test remote_process.run_process."""
    request = untrusted_runner_pb2.RunProcessRequest()
    request.cmdline = 'cmd arg'
    request.timeout = 20
    request.testcase_run = True
    request.gestures.extend(['gesture', 'gesture2'])
    request.env_copy.update({'VAR': 'VAL'})

    self.mock.run_process.return_value = (0, 10.0, 'output')

    response = remote_process.run_process(request, None)
    self.assertEqual(response.return_code, 0)
    self.assertEqual(response.execution_time, 10.0)
    self.assertEqual(response.output, 'output')

    self.mock.run_process.assert_called_with(
        cmdline='cmd arg',
        env_copy={'VAR': 'VAL'},
        gestures=['gesture', 'gesture2'],
        testcase_run=True,
        timeout=20)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Untrusted instance."""

from concurrent import futures
import functools
import os
import threading
import time
import traceback

import grpc

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.google_cloud_utils import compute_metadata
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.protos import heartbeat_pb2
from clusterfuzz._internal.protos import heartbeat_pb2_grpc
from clusterfuzz._internal.protos import untrusted_runner_pb2
from clusterfuzz._internal.protos import untrusted_runner_pb2_grpc
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import process_handler
from clusterfuzz._internal.system import shell

from . import build_setup
from . import config
from . import file_impl
from . import remote_process
from . import symbolize
from . import tasks_impl

SHUTDOWN_GRACE_SECONDS = 5


class WorkerState(object):
  """Worker's state."""

  def __init__(self):
    self.server = None
    self.shutting_down = threading.Event()
    self.start_time = None


_worker_state = WorkerState()

# Used to detect overlapping RPCs.
_rpc_count_lock = threading.Lock()
_rpc_count = 0


def wrap_servicer(func):
  """Wrap a servicer to add additional functionality."""

  @functools.wraps(func)
  def wrapper(self, request, context):  # pylint: disable=unused-argument
    """Wrapper function."""
    global _rpc_count

    # Check if there is a in-progress RPC.
    with _rpc_count_lock:
      if _rpc_count > 0:
        logs.log_error('Hung RPC detected, shutting down.')
        _worker_state.shutting_down.set()
        return None

      _rpc_count += 1

    try:
      result = func(self, request, context)
    except Exception:
      # Include full exception details.
      context.set_code(grpc.StatusCode.UNKNOWN)
      context.set_details(traceback.format_exc())
      raise
    finally:
      with _rpc_count_lock:
        assert _rpc_count > 0
        _rpc_count -= 1

    return result

  return wrapper


class UntrustedRunnerServicer(
    untrusted_runner_pb2_grpc.UntrustedRunnerServicer):
  """Untrusted runner implementation."""

  @wrap_servicer
  def GetStatus(self, request, context):  # pylint: disable=unused-argument
    return untrusted_runner_pb2.GetStatusResponse(
        revision=utils.current_source_version(),
        start_time=_worker_state.start_time,
        bot_name=environment.get_value('BOT_NAME'))

  @wrap_servicer
  def SetupRegularBuild(self, request, _):
    return build_setup.setup_regular_build(request)

  @wrap_servicer
  def RunAndWait(self, request, context):
    return remote_process.run_and_wait(request, context)

  @wrap_servicer
  def RunProcess(self, request, context):
    return remote_process.run_process(request, context)

  @wrap_servicer
  def CreateDirectory(self, request, context):
    return file_impl.create_directory(request, context)

  @wrap_servicer
  def RemoveDirectory(self, request, context):
    return file_impl.remove_directory(request, context)

  @wrap_servicer
  def ListFiles(self, request, context):
    return file_impl.list_files(request, context)

  @wrap_servicer
  def CopyFileTo(self, request_iterator, context):
    return file_impl.copy_file_to_worker(request_iterator, context)

  @wrap_servicer
  def CopyFileFrom(self, request, context):
    return file_impl.copy_file_from_worker(request, context)

  @wrap_servicer
  def Stat(self, request, context):
    return file_impl.stat(request, context)

  @wrap_servicer
  def UpdateEnvironment(self, request, _):
    os.environ.update(request.env)
    return untrusted_runner_pb2.UpdateEnvironmentResponse()

  @wrap_servicer
  def ResetEnvironment(self, _, context):  # pylint: disable=unused-argument
    environment.reset_environment()
    return untrusted_runner_pb2.ResetEnvironmentResponse()

  @wrap_servicer
  def UpdateSource(self, request, context):  # pylint: disable=unused-argument
    # Exit and let run.py update source.
    _worker_state.shutting_down.set()
    return untrusted_runner_pb2.UpdateSourceResponse()

  @wrap_servicer
  def SymbolizeStacktrace(self, request, _):
    return symbolize.symbolize_stacktrace(request)

  @wrap_servicer
  def TerminateStaleApplicationInstances(self, request, context):  # pylint: disable=unused-argument
    process_handler.terminate_stale_application_instances()
    return untrusted_runner_pb2.TerminateStaleApplicationInstancesResponse()

  @wrap_servicer
  def GetFuzzTargets(self, request, context):
    return file_impl.get_fuzz_targets(request, context)

  @wrap_servicer
  def ProcessTestcase(self, request, context):
    return tasks_impl.process_testcase(request, context)

  @wrap_servicer
  def EngineFuzz(self, request, context):
    return tasks_impl.engine_fuzz(request, context)

  @wrap_servicer
  def EngineReproduce(self, request, context):
    return tasks_impl.engine_reproduce(request, context)

  @wrap_servicer
  def PruneCorpus(self, request, context):
    return tasks_impl.prune_corpus(request, context)


class HeartbeatServicer(heartbeat_pb2_grpc.HeartbeatServicer):
  """Heartbeat service (for keeping connections alive)."""

  def Beat(self, _, context):  # pylint: disable=unused-argument
    return heartbeat_pb2.HeartbeatResponse()


def _get_tls_cert_and_key():
  """Get the TLS cert from instance metadata."""
  # TODO(ochang): Implement a fake metadata server for testing.
  local_cert_location = environment.get_value('UNTRUSTED_TLS_CERT_FOR_TESTING')
  local_key_location = environment.get_value('UNTRUSTED_TLS_KEY_FOR_TESTING')

  if local_cert_location and local_key_location:
    with open(local_cert_location, 'rb') as f:
      cert_contents = f.read()

    with open(local_key_location, 'rb') as f:
      key_contents = f.read()

    return cert_contents, key_contents

  cert_contents = compute_metadata.get('instance/attributes/tls-cert').encode()
  key_contents = compute_metadata.get('instance/attributes/tls-key').encode()
  return cert_contents, key_contents


def start_server():
  """Start the server."""
  # Check overall free disk space. If we are running too low, clear all
  # data directories like builds, fuzzers, data bundles, etc.
  shell.clear_data_directories_on_low_disk_space()

  cert_contents, key_contents = _get_tls_cert_and_key()
  assert cert_contents and key_contents
  server_credentials = grpc.ssl_server_credentials([(key_contents,
                                                     cert_contents)])
  _worker_state.server = grpc.server(
      futures.ThreadPoolExecutor(max_workers=config.NUM_WORKER_THREADS),
      options=config.GRPC_OPTIONS)

  untrusted_runner_pb2_grpc.add_UntrustedRunnerServicer_to_server(
      UntrustedRunnerServicer(), _worker_state.server)
  heartbeat_pb2_grpc.add_HeartbeatServicer_to_server(HeartbeatServicer(),
                                                     _worker_state.server)

  _worker_state.server.add_secure_port('[::]:%d' % config.PORT,
                                       server_credentials)

  _worker_state.start_time = int(time.time())
  _worker_state.server.start()

  logs.log('Server started.')

  # Run forever until shutdown.
  _worker_state.shutting_down.wait()

  logs.log('Server shutting down.')
  stopped = _worker_state.server.stop(SHUTDOWN_GRACE_SECONDS)
  stopped.wait()

  # Prevent python GIL deadlocks on shutdown. See https://crbug.com/744680.
  # pylint: disable=protected-access
  os._exit(0)


def server():
  """Return the grpc.Server."""
  return _worker_state.server



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for getting testcase variants."""
from clusterfuzz._internal.datastore import data_types
from handlers import base_handler
from libs import handler


class Handler(base_handler.Handler):
  """Handler that return testcase variants."""

  def get_variants(self, testcase):
    """Get testcase variants"""

    def _display_status(status):
      """Return status for display."""
      if status == data_types.TestcaseVariantStatus.PENDING:
        return 'Pending'
      if status == data_types.TestcaseVariantStatus.REPRODUCIBLE:
        return 'Reproducible'
      if status == data_types.TestcaseVariantStatus.FLAKY:
        return 'Flaky'
      if status == data_types.TestcaseVariantStatus.UNREPRODUCIBLE:
        return 'Unreproducible'

      return 'Unknown'

    items = []
    variants = data_types.TestcaseVariant.query(
        data_types.TestcaseVariant.testcase_id == testcase.key.id()).order(
            data_types.TestcaseVariant.job_type)
    for variant in variants:
      if variant.status == data_types.TestcaseVariantStatus.UNREPRODUCIBLE:
        # Avoid showing these to keep table small and minimize confusion with
        # some unrelated job types in the same project.
        continue

      is_pending = variant.status == data_types.TestcaseVariantStatus.PENDING
      item = {
          'isPending': is_pending,
          'status': _display_status(variant.status),
          'job': variant.job_type,
      }
      if not is_pending:
        item.update({
            'revision':
                variant.revision,
            'crashType':
                variant.crash_type,
            'crashStateLines': (variant.crash_state or '').strip().splitlines(),
            'securityFlag':
                variant.security_flag,
            'isSimilar':
                variant.is_similar,
            'reproducerKey':
                variant.reproducer_key,
        })
      items.append(item)

    return items

  @handler.get(handler.JSON)
  @handler.check_testcase_access
  def get(self, testcase):
    """Return testcase variants."""
    items = []
    message = None
    if testcase.one_time_crasher_flag:
      message = 'Not run for unreproducible testcases.'
    elif not testcase.minimized_keys:
      message = 'Pending, waiting for minimization to finish.'
    else:
      items = self.get_variants(testcase)

    response = {
        'items': items,
        'message': message,
    }
    return self.render_json(response)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Environment modification functions."""

import os
import re

import six

try:
  from clusterfuzz._internal.protos import untrusted_runner_pb2

  from . import file_host
  from . import host
except ImportError:
  # TODO(ochang): Fix this.
  pass

FORWARDED_ENVIRONMENT_VARIABLES = [
    re.compile(pattern) for pattern in (
        r'^AFL_.*',
        r'^APPLICATION_ID$',
        r'^ASAN_OPTIONS$',
        r'^BACKUP_BUCKET$',
        r'^CORPUS_BUCKET$',
        r'^MUTATOR_PLUGINS_BUCKET$',
        r'^FUZZ_CORPUS_DIR$',
        r'^FUZZER_DIR$',
        r'^FUZZER_NAME_REGEX$',
        r'^FUZZING_STRATEGIES$',
        r'^FUZZ_TARGET$',
        r'^FUZZ_TEST_TIMEOUT$',
        r'^GSUTIL_PATH$',
        r'^JOB_NAME$',
        r'^LOCAL_DEVELOPMENT$',
        r'^MSAN_OPTIONS$',
        r'^PATH$',
        r'^PY_UNITTESTS$',
        r'^QUARANTINE_BUCKET$',
        r'^SHARED_CORPUS_BUCKET$',
        r'^STRATEGY_SELECTION_DISTRIBUTION$',
        r'^STRATEGY_SELECTION_METHOD$',
        r'^TASK_NAME$',
        r'^TASK_PAYLOAD$',
        r'^TEST_TIMEOUT$',
        r'^TSAN_OPTIONS$',
        r'^UBSAN_OPTIONS$',
        r'^UNPACK_ALL_FUZZ_TARGETS_AND_FILES$',
        r'^USE_MINIJAIL$',
        r'^USER$',
    )
]

REBASED_ENVIRONMENT_VARIABLES = set([
    'FUZZER_DIR',
])


def is_forwarded_environment_variable(environment_variable):
  """Return whether or not |environment_variable| should be forwarded."""
  return any(
      pattern.match(environment_variable)
      for pattern in FORWARDED_ENVIRONMENT_VARIABLES)


def should_rebase_environment_value(environment_variable):
  """Return whether or not |environment_variable|'s value should be rebased."""
  return environment_variable in REBASED_ENVIRONMENT_VARIABLES


def update_environment(env):
  """Update worker's environment."""
  processed_env = {}
  for key, value in six.iteritems(env):
    if should_rebase_environment_value(key):
      value = file_host.rebase_to_worker_root(value)

    processed_env[key] = value

  request = untrusted_runner_pb2.UpdateEnvironmentRequest(env=processed_env)
  host.stub().UpdateEnvironment(request)


def set_environment_vars(env, source_env):
  """Copy allowed environment variables from |source_env|."""
  if not source_env:
    return

  for name, value in six.iteritems(source_env):
    if is_forwarded_environment_variable(name):
      # Avoid creating circular dependencies from importing environment by
      # using os.getenv.
      if os.getenv('TRUSTED_HOST') and should_rebase_environment_value(name):
        value = file_host.rebase_to_worker_root(value)

      env[name] = value


def get_env_for_untrusted_process(overrides):
  """Return environment for running an untrusted process."""
  env = {}
  if overrides is not None:
    set_environment_vars(env, overrides)
  else:
    set_environment_vars(env, os.environ)
  return env


def forward_environment_variable(key, value):
  """Forward the environment variable if needed."""
  if not host.is_initialized():
    return

  if is_forwarded_environment_variable(key):
    update_environment({key: value})


def reset_environment():
  """Reset environment variables."""
  request = untrusted_runner_pb2.ResetEnvironmentRequest()
  host.stub().ResetEnvironment(request)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""External tasks."""

from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.google_cloud_utils import blobs
from clusterfuzz._internal.google_cloud_utils import pubsub
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment


def add_external_task(command, testcase_id, job):
  """Add external task."""
  if command != 'progression':
    # Only progression is supported.
    return

  pubsub_client = pubsub.PubSubClient()
  topic_name = job.external_reproduction_topic
  assert topic_name is not None

  testcase = data_handler.get_testcase_by_id(testcase_id)
  fuzz_target = testcase.get_fuzz_target()

  memory_tool_name = environment.get_memory_tool_name(job.name)
  sanitizer = environment.SANITIZER_NAME_MAP.get(memory_tool_name)
  job_environment = job.get_environment()
  if job_environment.get('CUSTOM_BINARY'):
    raise RuntimeError('External jobs should never have custom binaries.')

  build_path = (
      job_environment.get('RELEASE_BUILD_BUCKET_PATH') or
      job_environment.get('FUZZ_TARGET_BUILD_BUCKET_PATH'))
  if build_path is None:
    raise RuntimeError(f'{job.name} has no build path defined.')

  min_revision = (
      testcase.get_metadata('last_tested_revision') or testcase.crash_revision)

  logs.log(f'Publishing external reproduction task for {testcase_id}.')
  attributes = {
      'project': job.project,
      'target': fuzz_target.binary,
      'fuzzer': testcase.fuzzer_name,
      'sanitizer': sanitizer,
      'job': job.name,
      'testcaseId': str(testcase_id),
      'buildPath': build_path,
      'minRevisionAbove': str(min_revision),
  }

  reproducer = blobs.read_key(testcase.fuzzed_keys)
  message = pubsub.Message(data=reproducer, attributes=attributes)
  pubsub_client.publish(topic_name, [message])



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Cron job to get the latest code coverage stats and HTML reports."""

import datetime
import json
import os

from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from handlers import base_handler
from libs import handler


def _latest_report_info_dir(bucket):
  """Returns a GCS URL to the latest report info for the given bucket."""
  return 'gs://{0}/latest_report_info/'.format(bucket)


def _basename(gcs_path):
  """Returns the basename for the given path without file extension."""
  return os.path.splitext(os.path.basename(gcs_path))[0]


def _read_json(url):
  """Returns a JSON obejct loaded from the given GCS url."""
  data = storage.read_data(url)

  result = None
  try:
    result = json.loads(data)
  except Exception as e:
    logs.log_warn(
        'Empty or malformed code coverage JSON (%s): %s.' % (url, str(e)))

  return result


def _coverage_information(summary_path, name, report_info):
  """Returns a CoverageInformation entity with coverage stats populated."""
  date = datetime.datetime.strptime(
      report_info['report_date'],
      data_types.COVERAGE_INFORMATION_DATE_FORMAT).date()

  # |name| can be either a project qualified fuzz target name or a project name.
  cov_info = data_handler.get_coverage_information(
      name, date, create_if_needed=True)
  cov_info.fuzzer = name
  cov_info.date = date

  # Link to a per project report as long as we don't have per fuzzer reports.
  cov_info.html_report_url = report_info['html_report_url']

  summary = _read_json(summary_path)
  if not summary:
    # We can encounter empty JSON files for broken fuzz targets.
    return cov_info

  total_stats = summary['data'][0]['totals']
  cov_info.functions_covered = total_stats['functions']['covered']
  cov_info.functions_total = total_stats['functions']['count']
  cov_info.edges_covered = total_stats['regions']['covered']
  cov_info.edges_total = total_stats['regions']['count']

  return cov_info


def _process_fuzzer_stats(fuzzer, project_info, project_name, bucket):
  """Processes coverage stats for a single fuzz target."""
  fuzzer_name = data_types.fuzz_target_project_qualified_name(
      project_name, _basename(fuzzer))
  fuzzer_info_path = storage.get_cloud_storage_file_path(bucket, fuzzer)
  logs.log(
      'Processing fuzzer stats for %s (%s).' % (fuzzer_name, fuzzer_info_path))
  return _coverage_information(fuzzer_info_path, fuzzer_name, project_info)


def _process_project_stats(project_info, project_name):
  """Processes coverage stats for a single project."""
  summary_path = project_info['report_summary_path']
  logs.log('Processing total stats for %s project (%s).' % (project_name,
                                                            summary_path))
  return _coverage_information(summary_path, project_name, project_info)


def _process_project(project, bucket):
  """Collects coverage information for all fuzz targets in the given project and
  the total stats for the project."""
  project_name = _basename(project)
  logs.log('Processing coverage for %s project.' % project_name)
  report_path = storage.get_cloud_storage_file_path(bucket, project)
  report_info = _read_json(report_path)
  if not report_info:
    logs.log_warn('Skipping code coverage for %s project.' % project_name)
    return

  # Iterate through report_info['fuzzer_stats_dir'] and prepare
  # CoverageInformation entities for invididual fuzz targets.
  entities = []
  for fuzzer in storage.list_blobs(
      report_info['fuzzer_stats_dir'], recursive=False):
    entities.append(
        _process_fuzzer_stats(fuzzer, report_info, project_name, bucket))

  logs.log('Processed coverage for %d targets in %s project.' % (len(entities),
                                                                 project_name))

  # Prepare CoverageInformation entity for the total project stats.
  entities.append(_process_project_stats(report_info, project_name))

  ndb_utils.put_multi(entities)


def collect_fuzzer_coverage(bucket):
  """Actual implementation of the fuzzer coverage task."""
  url = _latest_report_info_dir(bucket)
  for project in storage.list_blobs(url, recursive=False):
    _process_project(project, bucket)


class Handler(base_handler.Handler):
  """Collects the latest code coverage stats and links to reports."""

  @handler.cron()
  def get(self):
    """Handle a GET request."""
    # The task is supposed to be super reliable and never fail. If anything goes
    # wrong, we just fail with the exception going straight into StackDriver.
    logs.log('FuzzerCoverage task started.')
    bucket = local_config.ProjectConfig().get('coverage.reports.bucket')
    if not bucket:
      logs.log(
          'Coverage bucket is not specified. Skipping FuzzerCoverage task.')
      return

    collect_fuzzer_coverage(bucket)
    logs.log('FuzzerCoverage task finished successfully.')



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions for managing corpus tags."""

from clusterfuzz._internal.datastore import data_types


def get_fuzz_target_tags(fully_qualified_fuzz_target_name):
  """Get all the tags of a given fuzz target."""
  query = data_types.CorpusTag().query()
  query = query.filter(data_types.CorpusTag.fully_qualified_fuzz_target_name ==
                       fully_qualified_fuzz_target_name)
  return query.fetch()


def get_targets_with_tag(tag):
  """Get all fuzz targets with a given tag."""
  query = data_types.CorpusTag().query()
  query = query.filter(data_types.CorpusTag.tag == tag)
  return query.fetch()


def get_similarly_tagged_fuzzers(current_fully_qualified_fuzzer_name):
  """Get all fuzz targets with any similar tags to current fuzzer."""
  similarly_tagged_targets = {}
  for tag in get_fuzz_target_tags(current_fully_qualified_fuzzer_name):
    targets = [
        target.fully_qualified_fuzz_target_name
        for target in get_targets_with_tag(tag.tag)
        if target.fully_qualified_fuzz_target_name !=
        current_fully_qualified_fuzzer_name
    ]
    if targets:
      similarly_tagged_targets[tag.tag] = targets

  return similarly_tagged_targets



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration helpers for the reproduce tool."""

from urllib import parse

from clusterfuzz._internal.base import json_utils
from local.butler.reproduce_tool import errors
from local.butler.reproduce_tool import http_utils

REPRODUCE_TOOL_CONFIG_HANDLER = '/reproduce-tool/get-config'


class ReproduceToolConfiguration(object):
  """Dynamically loaded configuration for the reproduce tool."""

  def __init__(self, testcase_url):
    testcase_url_parts = parse.urlparse(testcase_url)
    config_url = testcase_url_parts._replace(
        path=REPRODUCE_TOOL_CONFIG_HANDLER).geturl()
    response, content = http_utils.request(config_url, body={})
    if response.status != 200:
      raise errors.ReproduceToolUnrecoverableError('Failed to access server.')

    self._config = json_utils.loads(content)

  def get(self, key):
    """Get a config entry."""
    return self._config[key]



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for batch_fuzzer_jobs."""

import unittest

from google.cloud import ndb

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import batch_fuzzer_jobs


@test_utils.with_cloud_emulators('datastore')
class TestBatchingFuzzerJobs(unittest.TestCase):
  """Test batching FuzzerJob entitites."""

  def setUp(self):
    self.total_fuzzer_jobs = 7000
    self.platforms = ['LINUX', 'WINDOWS']
    fuzzer_jobs = []
    for platform in self.platforms:
      for i in range(self.total_fuzzer_jobs):
        fuzzer_job = data_types.FuzzerJob(
            fuzzer='libFuzzer',
            job='libfuzzer_asan_{}_{:06d}'.format(platform, i),
            platform=platform)
        fuzzer_jobs.append(fuzzer_job)

    ndb.put_multi(fuzzer_jobs)

    # Should be removed.
    data_types.FuzzerJobs(id='LINUX-2', platform='LINUX').put()

    # Should be overwritten and not removed.
    data_types.FuzzerJobs(id='LINUX-0', platform='LINUX').put()

  def test_batch(self):
    """Test batching."""
    batch_fuzzer_jobs.batch_fuzzer_jobs()
    for platform in self.platforms:
      all_fuzzer_jobs = []
      for i in range(2):
        key = ndb.Key(data_types.FuzzerJobs, platform + '-' + str(i))
        fuzzer_jobs = key.get()
        all_fuzzer_jobs.extend(fuzzer_jobs.fuzzer_jobs)

      self.assertEqual(self.total_fuzzer_jobs, len(all_fuzzer_jobs))
      for i in range(self.total_fuzzer_jobs):
        self.assertEqual('libfuzzer_asan_{}_{:06d}'.format(platform, i),
                         all_fuzzer_jobs[i].job)

    self.assertIsNone(ndb.Key(data_types.FuzzerJobs, 'LINUX-2').get())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the HTML minimizer."""

import unittest

from clusterfuzz._internal.bot.minimizer import html_minimizer
from clusterfuzz._internal.tests.test_libs import helpers


class HTMLMinimizerTest(unittest.TestCase):
  """Test for HTML Minimizer."""

  def setUp(self):
    helpers.patch_environ(self)
    helpers.patch(self, [(
        'js_minimizer',
        'clusterfuzz._internal.bot.minimizer.js_minimizer.JSMinimizer.minimize'
    ), (
        'html_minimizer',
        'clusterfuzz._internal.bot.minimizer.chunk_minimizer.ChunkMinimizer.minimize'
    ), ('line_minimizer',
        'clusterfuzz._internal.bot.minimizer.delta_minimizer.DeltaMinimizer.minimize'
       )])

    # HTML Minimizer passes data to a series of sub-minimizers. These take a
    # long time to run. We just want to test that the results are being handled
    # correctly.
    self.mock.js_minimizer.side_effect = self._mock_js_minimization
    self.mock.html_minimizer.side_effect = self._mock_html_minimization
    self.mock.line_minimizer.side_effect = self._mock_line_minimization

    self._html_data_begin = b"""
    <!-- /mnt/fuzzer/gecko-tests/layout/reftests/bugs/256180-2.html -->
    <!DOCTYPE html>
    <html>
    <head>
    <meta charset="utf-8">
    <font></font>
    <script>"""

    self._script_data = b"""
    function a() {
        var a = 0;
    }
    
    function b() {
        var b = document.createTextNode("PASS");
    }"""

    self._html_data_end = b"""</script>
    <style>
        div {
            display: table-cell;
        }
    </style>
    </html>"""

    self._simple_test = (
        self._html_data_begin + self._script_data + self._html_data_end)

    self._minimizer = html_minimizer.HTMLMinimizer(self._mock_test_function())

  def _mock_test_function(self):
    return True

  def _mock_js_minimization(self, minimizer, data):
    """Mock js_minimization. Returns simple js_minimized code."""
    # pylint: disable=unused-argument
    return minimizer.token_combiner(
        [b'var b = document.createTextNode("PASS")\n'])

  def _mock_html_minimization(self, minimizer, data):
    """Mock html minimization. Returns minimized top tags when its called for
    the first section, otherwise returns minimized bottom section"""
    # The first 3 times the html_minimizer is called will be for the begining.
    if b'<html>' in data:
      return minimizer.token_combiner([b'<html>\n<script>\n'])

    return minimizer.token_combiner([b'</script>\n</html>'])

  def _mock_line_minimization(self, minimizer, data):
    """Mock line minimization. Assume data is already line-minimized."""
    # pylint: disable=unused-argument
    return data

  def test_get_tokens_and_metadata_splits_correct_number(self):
    """Test that the html minimizer splits the data into the correct number of
    sections."""
    tokens = self._minimizer.get_tokens_and_metadata(self._simple_test)

    self.assertEqual(len(tokens), 3)

  def test_get_tokens_and_metadata_splits_correct_types(self):
    """Test that each of the sections the data is split into for minimization
    are classified correctly."""
    tokens = self._minimizer.get_tokens_and_metadata(self._simple_test)

    self.assertEqual(tokens[0].token_type, self._minimizer.Token.TYPE_HTML)
    self.assertEqual(tokens[1].token_type, self._minimizer.Token.TYPE_SCRIPT)
    self.assertEqual(tokens[2].token_type, self._minimizer.Token.TYPE_HTML)

  def test_get_tokens_and_metadata_splits_correct_data(self):
    """Test that each of the sections the data is split into for minimization
    contains the correct data"""
    tokens = self._minimizer.get_tokens_and_metadata(self._simple_test)

    self.assertEqual(tokens[0].data, self._html_data_begin)
    self.assertEqual(tokens[1].data, self._script_data)
    self.assertEqual(tokens[2].data, self._html_data_end)

  def test_minimization_returns_correct_result(self):
    """Test that the minimizer goes through all of the correct sub-minimizers
    gives the correct tokenizers, and returns the minimized product."""
    minimized = b'<html>\n<script>\nvar b = document.createTextNode("PASS")' \
                b'\n</script>\n</html>'
    res = self._minimizer.minimize(self._simple_test)
    self.assertEqual(res, minimized)

  def test_combine_worker_tokens_with_prefix_and_suffix(self):
    """Test that combine worker tokens works the way that the minimizer expects
    it to."""
    prefix = b'PRE'
    suffix = b'POST'
    tokens = [b'Here', b'Are', b'The', b'Tokens']

    combo = self._minimizer.combine_worker_tokens(tokens, prefix, suffix)

    self.assertEqual(combo, b'PREHereAreTheTokensPOST')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Reproduce tool configuration tests."""
import unittest

from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs.reproduce_tool_fakes import \
    FakeResponse
from local.butler.reproduce_tool import config
from local.butler.reproduce_tool import errors


class ReproduceToolConfigurationTest(unittest.TestCase):
  """Tests for ReproduceToolConfiguration."""

  def setUp(self):
    helpers.patch(self, [
        'local.butler.reproduce_tool.http_utils.request',
    ])

  def test_local_server(self):
    """Ensure that we can get the configuration for a local server."""
    self.mock.request.return_value = (FakeResponse(200), '{"x": 1}')
    configuration = config.ReproduceToolConfiguration(
        'http://localhost:9000/testcase-detail/1')

    self.assertEqual(configuration.get('x'), 1)
    self.mock.request.assert_called_once_with(
        'http://localhost:9000/reproduce-tool/get-config', body={})

  def test_https_server(self):
    """Ensure that we can get the configuration for an HTTPS server."""
    self.mock.request.return_value = (FakeResponse(200), '{"x": 1}')
    configuration = config.ReproduceToolConfiguration(
        'https://clusterfuzz/testcase-detail/1')

    self.assertEqual(configuration.get('x'), 1)
    self.mock.request.assert_called_once_with(
        'https://clusterfuzz/reproduce-tool/get-config', body={})

  def test_failure(self):
    """Ensure that we raise an exception if the server encounters an error."""
    self.mock.request.return_value = (FakeResponse(403), '{"x": 1}')

    with self.assertRaises(errors.ReproduceToolUnrecoverableError):
      config.ReproduceToolConfiguration('https://clusterfuzz/testcase-detail/1')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Constants that are meaningful to AFL.
Should not have any dependancies.
"""

# AFL flags.
DICT_FLAG = '-x'

INPUT_FLAG = '-i'

MEMORY_LIMIT_FLAG = '-m'

OUTPUT_FLAG = '-o'

TIMEOUT_FLAG = '-t'

INSTANCE_ID_FLAG = '-S'

MOPT_FLAG = '-L'

CMPLOG_LEVEL_FLAG = '-l'

QUEUE_OLD_STRATEGY_FLAG = '-Z'

SCHEDULER_FLAG = '-p'

CMPLOG_FLAG = '-c'

# AFL CMPLOG suboptions.
CMPLOG_ARITH = 'A'

CMPLOG_TRANS = 'T'

# AFL environment variables.
AFL_MAP_SIZE_ENV_VAR = 'AFL_MAP_SIZE'

IGNORE_UNKNOWN_ENVS_ENV_VAR = 'AFL_IGNORE_UNKNOWN_ENVS'

SKIP_CRASHES_ENV_VAR = 'AFL_SKIP_CRASHES'

SKIP_CPUFREQ_ENV_VAR = 'AFL_SKIP_CPUFREQ'

BENCH_UNTIL_CRASH_ENV_VAR = 'AFL_BENCH_UNTIL_CRASH'

DONT_DEFER_ENV_VAR = 'AFL_DRIVER_DONT_DEFER'

FAST_CAL_ENV_VAR = 'AFL_FAST_CAL'

FORKSRV_INIT_TMOUT_ENV_VAR = 'AFL_FORKSRV_INIT_TMOUT'

NO_AFFINITY_ENV_VAR = 'AFL_NO_AFFINITY'

STDERR_FILENAME_ENV_VAR = 'AFL_DRIVER_STDERR_DUPLICATE_FILENAME'

CLOSE_FD_MASK_ENV_VAR = 'AFL_DRIVER_CLOSE_FD_MASK'

EXPAND_HAVOC_NOW_VAR = 'AFL_EXPAND_HAVOC_NOW'

DEBUG_VAR = 'AFL_DEBUG'

CMPLOG_ONLY_NEW_ENV_VAR = 'AFL_CMPLOG_ONLY_NEW'

DISABLE_TRIM_ENV_VAR = 'AFL_DISABLE_TRIM'

EXPAND_HAVOC_ENV_VAR = 'AFL_EXPAND_HAVOC_NOW'

# Other settings.

MAX_FILE_BYTES = 2**20  # 1 MB

FORKSERVER_TIMEOUT = 60000  # milliseconds

# This should be as high as possible, otherwise AFL will restart the binary too
# often to be competitive with pure libFuzzer. 2147483647 is the maximum signed
# integer. afl_driver accepts one argument which it converts to a signed int
# using atoi hence this is the largest value we can pick.
MAX_PERSISTENT_EXECUTIONS = '2147483647'

# Resume is used by passing -i- to AFL. See https://goo.gl/rZi455
RESUME_INPUT = '-'

# Don't let afl set a memory limit. Otherwise we will not be able to use
# sanitizers.
MAX_MEMORY_LIMIT = 'none'

CORE_PATTERN_FILE_PATH = '/proc/sys/kernel/core_pattern'

DEFAULT_INSTANCE_ID = 'default'



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: clusterfuzz/_internal/protos/heartbeat.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='clusterfuzz/_internal/protos/heartbeat.proto',
  package='',
  syntax='proto2',
  serialized_options=b'Z#clusterfuzz/protos/untrusted_runner',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n,clusterfuzz/_internal/protos/heartbeat.proto\"\x12\n\x10HeartbeatRequest\"\x13\n\x11HeartbeatResponse2:\n\tHeartbeat\x12-\n\x04\x42\x65\x61t\x12\x11.HeartbeatRequest\x1a\x12.HeartbeatResponseB%Z#clusterfuzz/protos/untrusted_runner'
)




_HEARTBEATREQUEST = _descriptor.Descriptor(
  name='HeartbeatRequest',
  full_name='HeartbeatRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=48,
  serialized_end=66,
)


_HEARTBEATRESPONSE = _descriptor.Descriptor(
  name='HeartbeatResponse',
  full_name='HeartbeatResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=68,
  serialized_end=87,
)

DESCRIPTOR.message_types_by_name['HeartbeatRequest'] = _HEARTBEATREQUEST
DESCRIPTOR.message_types_by_name['HeartbeatResponse'] = _HEARTBEATRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

HeartbeatRequest = _reflection.GeneratedProtocolMessageType('HeartbeatRequest', (_message.Message,), {
  'DESCRIPTOR' : _HEARTBEATREQUEST,
  '__module__' : 'clusterfuzz._internal.protos.heartbeat_pb2'
  # @@protoc_insertion_point(class_scope:HeartbeatRequest)
  })
_sym_db.RegisterMessage(HeartbeatRequest)

HeartbeatResponse = _reflection.GeneratedProtocolMessageType('HeartbeatResponse', (_message.Message,), {
  'DESCRIPTOR' : _HEARTBEATRESPONSE,
  '__module__' : 'clusterfuzz._internal.protos.heartbeat_pb2'
  # @@protoc_insertion_point(class_scope:HeartbeatResponse)
  })
_sym_db.RegisterMessage(HeartbeatResponse)


DESCRIPTOR._options = None

_HEARTBEAT = _descriptor.ServiceDescriptor(
  name='Heartbeat',
  full_name='Heartbeat',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=89,
  serialized_end=147,
  methods=[
  _descriptor.MethodDescriptor(
    name='Beat',
    full_name='Heartbeat.Beat',
    index=0,
    containing_service=None,
    input_type=_HEARTBEATREQUEST,
    output_type=_HEARTBEATRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_HEARTBEAT)

DESCRIPTOR.services_by_name['Heartbeat'] = _HEARTBEAT

# @@protoc_insertion_point(module_scope)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for ml rnn model."""

import os
import random
import sys

import numpy as np
import tensorflow as tf

from clusterfuzz._internal.bot.fuzzers.ml.rnn import constants


def build_model(num_rnn_cells, dropout_pkeep, batch_size, debug):
  """Build the RNN model.

  Since we use the Keras sequential model and we use different batch sizes for
  train, validation and demo output generation, we use this function to rebatch
  the model.

  Args:
    num_rnn_cells: number of RNN cells to use.
    dropout_pkeep: probability of keeping a node in dropout.
    batch_size: batch size used by the model layer.
    debug: if True, print a summary of the model.

  Returns:
    Keras Sequential RNN model.
  """
  dropout_pdrop = 1 - dropout_pkeep
  model = tf.keras.Sequential([
      tf.keras.layers.Embedding(
          constants.ALPHA_SIZE,
          constants.ALPHA_SIZE,
          batch_input_shape=[batch_size, None]),
      tf.keras.layers.GRU(
          num_rnn_cells,
          return_sequences=True,
          stateful=True,
          dropout=dropout_pdrop),
      tf.keras.layers.Dense(constants.ALPHA_SIZE),
  ])

  # Display a summary of the model to debug shapes.
  if debug:
    model.summary()

  return model


def validate_model_path(model_path):
  """RNN model consists of two files. This validates if they all exist."""
  model_exists = (
      os.path.exists(model_path + constants.MODEL_DATA_SUFFIX) and
      os.path.exists(model_path + constants.MODEL_INDEX_SUFFIX))
  return model_exists


def random_element_from_list(sequence):
  """Returns a random element from a list."""
  return random.SystemRandom().choice(sequence)


def get_files_list(directory):
  """Returns a list of files in a directory (recursively)."""
  files_list = []
  for (root, _, files) in os.walk(directory):
    for filename in files:
      file_path = os.path.join(root, filename)
      if not os.path.isfile(file_path):
        continue
      files_list.append(file_path)

  return files_list


def get_files_info(directory):
  """Returns a list of files info in a directory (recursively).

  Args:
    directory: Directory path.

  Returns:
    A list of file info. Each element is a dictionary with keys 'first_byte'
    and 'file_length'. Note that the list excludes empty files (zero length).
  """
  files_list = get_files_list(directory)
  files_info_list = []

  # Get files info. Ignore empty files.
  for file_path in files_list:
    file_size = os.path.getsize(file_path)
    if not file_size:
      continue
    with open(file_path, 'rb') as f:
      first_byte = ord(f.read(1))
      files_info_list.append({'first_byte': first_byte, 'file_size': file_size})

  return files_info_list


def encode_text(bytes_data):
  """Encode byte string to a list of integers.

  Args:
    bytes_data: Byte string.

  Returns:
    An encoded list of integers representing code points.
  """
  return list(bytes_data)


def decode_to_text(encoded_list):
  """Decode an encoded list.

  Args:
    encoded_list: Encoded list of code points.

  Returns:
    A string of decoded data.
  """
  return ''.join([chr(c) for c in encoded_list])


def sample_from_probabilities(probabilities, topn=constants.ALPHA_SIZE):
  """Randomly choose one byte from topn bytes based on their probabilities.

  Roll the dice to produce a random integer in the [0..ALPHA_SIZE] range,
  according to the provided probabilities. If topn is specified, only the
  topn highest probabilities are taken into account.

  Args:
    probabilities: A list of size ALPHA_SIZE with individual probabilities.
    topn: The number of highest probabilities to consider. Defaults to all of
        them.

  Returns:
    A random integer.
  """
  p = np.squeeze(probabilities)
  p[np.argsort(p)[:-topn]] = 0
  p = p / np.sum(p)
  return np.random.choice(constants.ALPHA_SIZE, 1, p=p)[0]


def rnn_minibatch_sequencer(raw_data, batch_size, sequence_size, nb_epochs):
  """Divide data into batches and return one batch for training each time.

  Divide the data into batches of sequences so that all the sequences
  in one batch continue in the next batch. This is a generator that will
  keep returning batches until the input data has been seen nb_epochs times.
  Sequences are continued even between epochs, apart from one,
  the one corresponding to the end of raw_data.

  The remainder at the end of raw_data that does not fit in an full batch is
  ignored.

  Args:
    raw_data: The training text.
    batch_size: The size of a training minibatch.
    sequence_size: The unroll size of the RNN.
    nb_epochs: Number of epochs to train on.

  Yields:
    x: One batch of training sequences.
    y: One batch of target sequences, i.e. training sequences shifted by 1.
    epoch: The current epoch number (starting at 0).
  """
  data = np.array(raw_data)
  data_len = data.shape[0]

  # Using (data_len-1) because we must provide for the sequence
  # shifted by 1 too.
  nb_batches = (data_len - 1) // (batch_size * sequence_size)
  assert nb_batches > 0, ('Not enough data, even for a single batch. Try using '
                          'a smaller batch_size.')
  rounded_data_len = nb_batches * batch_size * sequence_size
  xdata = np.reshape(data[0:rounded_data_len],
                     [batch_size, nb_batches * sequence_size])
  ydata = np.reshape(data[1:rounded_data_len + 1],
                     [batch_size, nb_batches * sequence_size])

  for epoch in range(nb_epochs):
    for batch in range(nb_batches):
      x = xdata[:, batch * sequence_size:(batch + 1) * sequence_size]
      y = ydata[:, batch * sequence_size:(batch + 1) * sequence_size]

      # To continue the text from epoch to epoch (do not reset rnn state!).
      x = np.roll(x, -epoch, axis=0)
      y = np.roll(y, -epoch, axis=0)
      yield x, y, epoch


def find_input(index, input_ranges):
  """Find the input name given the index of training data."""
  return next(input['name']
              for input in input_ranges
              if input['start'] <= index < input['end'])


def find_input_index(index, input_ranges):
  """Find the input index given the index of training data."""
  return next(i for i, input in enumerate(input_ranges)
              if input['start'] <= index < input['end'])


def print_learning_learned_comparison(x, y, losses, input_ranges, batch_loss,
                                      batch_accuracy, epoch_size, index, epoch):
  """Display utility for printing learning statistics."""
  print()

  # epoch_size in number of batches.
  batch_size = x.shape[0]
  sequence_len = x.shape[1]
  start_index_in_epoch = index % (epoch_size * batch_size * sequence_len)
  for k in range(batch_size):
    index_in_epoch = index % (epoch_size * batch_size * sequence_len)
    decx = repr(decode_to_text(x[k]))
    decy = repr(decode_to_text(y[k]))
    formatted_decx = '{: <40.40}'.format(decx)
    formatted_decy = '{: <40.40}'.format(decy)
    inputname = find_input(index_in_epoch, input_ranges)
    formatted_inputname = '{: <10.20}'.format(inputname)
    epoch_string = '{:6d}'.format(index) + ' (epoch {}) '.format(epoch)
    loss_string = 'loss: {:.5f}'.format(losses[k])
    print_string = epoch_string + formatted_inputname + ' | {} | {} | {}'
    print(print_string.format(formatted_decx, formatted_decy, loss_string))
    index += sequence_len

  format_string = '{:-^' + str(len(epoch_string)) + '}'
  format_string += '{:-^' + str(len(formatted_inputname)) + '}'
  format_string += '{:-^' + str(len(formatted_decx) + 4) + '}'
  format_string += '{:-^' + str(len(formatted_decy) + 4) + '}'
  format_string += '{:-^' + str(len(loss_string)) + '}'
  footer = format_string.format('INDEX', 'INPUT NAME',
                                'TRAINING SEQUENCE (truncated)',
                                'PREDICTED SEQUENCE (truncated)', 'LOSS')
  print(footer)

  # Print statistics
  batch_index = start_index_in_epoch // (batch_size * sequence_len)
  batch_string = 'batch {}/{} in epoch {},'.format(batch_index, epoch_size,
                                                   epoch)
  stats = '{: <28} batch loss: {:.5f}, batch accuracy: {:.5f}'.format(
      batch_string, batch_loss, batch_accuracy)
  print('\nTRAINING STATS: {}'.format(stats))


class Progress(object):
  """Text mode progress bar.

  Usage:
    p = Progress(30)
    p.step()
    p.step()
    p.step(start=True) # to restart form 0%

  The progress bar displays a new header at each restart.
  """

  def __init__(self, maxi, size=100, msg=''):
    """Initialize class.

    Args:
      maxi: The number of steps required to reach 100%.
      size: The number of characters taken on the screen by the progress bar.
      msg: The message displayed in the header of the progress bar.
    """
    self.maxi = maxi

    self.p = self.__start_progress(maxi)()
    self.header_printed = False
    self.msg = msg
    self.size = size

  def step(self, reset=False):
    """Print one step of progress bar."""
    if reset:
      self.__init__(self.maxi, self.size, self.msg)
    if not self.header_printed:
      self.__print_header()
    next(self.p)

  def __print_header(self):
    """Print progress bar header."""
    format_string = '\n0%{: ^' + str(self.size - 6) + '}100%'
    print(format_string.format(self.msg))
    self.header_printed = True

  def __start_progress(self, maxi):
    """Progress bar printer."""

    def print_progress():
      """Yields the number of dots printed."""
      # Bresenham's algorithm. Yields the number of dots printed.
      # This will always print 100 dots in max invocations.
      dx = maxi
      dy = self.size
      d = dy - dx
      for _ in range(maxi):
        k = 0
        while d >= 0:
          print('=', end='')
          sys.stdout.write('')
          sys.stdout.flush()
          k += 1
          d -= dx
        d += dy
        yield k

    return print_progress


def read_data_files(directory, validation=False):
  """Read data files (recursively) and split to training and validation sets.

  Optionally set aside the last file as validation data. No validation
  data is returned if there are 5 files or less.

  Args:
    directory: Directory path.
    validation: If True, sets the last file aside as validation data.

  Returns:
    Training data, validation data, list of loaded file names with ranges
    if validation is true.
  """
  code_text = []
  input_ranges = []
  input_list = get_files_list(directory)
  for input_file in input_list:
    with open(input_file, 'rb') as f:
      start = len(code_text)
      code_text.extend(encode_text(f.read()))
      end = len(code_text)
      input_ranges.append({
          'start': start,
          'end': end,
          'name': input_file.rsplit('/', 1)[-1]
      })

  print('Loaded {} corpus files.'.format(len(input_list)))

  if not input_ranges:
    sys.exit('No training data has been found. Aborting.')

  # For validation, use roughly 90K of text,
  # but no more than 10% of the entire text
  # and no more than 1 input in 5 => no validation at all for 5 files or fewer.

  # 10% of the text is how many files?
  total_len = len(code_text)
  validation_len = 0
  nb_inputs1 = 0
  for one_input in reversed(input_ranges):
    validation_len += one_input['end'] - one_input['start']
    nb_inputs1 += 1
    if validation_len > total_len // 10:
      break

  # 90K of text is how many inputs?
  validation_len = 0
  nb_inputs2 = 0
  for one_input in reversed(input_ranges):
    validation_len += one_input['end'] - one_input['start']
    nb_inputs2 += 1
    if validation_len > 90 * 1024:
      break

  # 20% of the inputs is how many inputs?
  nb_inputs3 = len(input_ranges) // 5

  # Pick the smallest.
  nb_inputs = min(nb_inputs1, nb_inputs2, nb_inputs3)

  if nb_inputs == 0 or not validation:
    cutoff = len(code_text)
  else:
    cutoff = input_ranges[-nb_inputs]['start']
  validation_text = code_text[cutoff:]
  code_text = code_text[:cutoff]
  return code_text, validation_text, input_ranges


def print_data_stats(data_len, validation_len, epoch_size):
  """Print training data statistics, such as size, batches."""
  data_len_mb = data_len / 1024.0 / 1024.0
  validation_len_kb = validation_len / 1024.0
  print('Training text size is {:.2f}MB with {:.2f}KB set aside for validation.'
        .format(data_len_mb, validation_len_kb) +
        'There will be {} batches per epoch'.format(epoch_size))


def print_validation_header(validation_start, input_ranges):
  """Print validation header."""
  input_index = find_input_index(validation_start, input_ranges)
  inputs = ''
  for i in range(input_index, len(input_ranges)):
    inputs += input_ranges[i]['name']
    if i < len(input_ranges) - 1:
      inputs += ', '


def print_validation_stats(loss, accuracy):
  """Print validation results, including loss and accuracy."""
  print('VALIDATION STATS:                                  ' +
        'loss: {:.5f},       accuracy: {:.5f}'.format(loss, accuracy))


def print_text_generation_header():
  """Print generation header."""
  print('\n{:-^138}'.format(' Generating random text from learned state '))


def print_text_generation_footer():
  """Print generation footer."""
  print('\n{:-^138}'.format(' End of generation '))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""GCS redirector."""

from flask import request

from clusterfuzz._internal.google_cloud_utils import storage
from handlers import base_handler
from libs import handler
from libs import helpers


class Handler(base_handler.Handler):
  """Gcs redirector."""

  @handler.get(handler.HTML)
  def get(self):
    """Handle a get request."""
    gcs_path = request.args.get('path', '')
    if not gcs_path:
      raise helpers.EarlyExitException('No path provided.', 400)

    if storage.get(gcs_path):
      host_url = storage.OBJECT_URL
    else:
      host_url = storage.DIRECTORY_URL

    bucket_name, object_path = storage.get_bucket_name_and_path(gcs_path)
    return self.redirect(host_url + '/' + bucket_name + '/' + object_path)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""crash_stats tests."""
import unittest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.metrics import crash_stats
from clusterfuzz._internal.tests.test_libs import test_utils


@test_utils.with_cloud_emulators('datastore')
class GetLastSuccessfulHourTest(unittest.TestCase):
  """Test get_last_successful_hour."""

  def test_none(self):
    """Get none because there's no last hour."""
    self.assertIsNone(crash_stats.get_last_successful_hour())

  def test_last_hour(self):
    """Get the last hour."""
    hour2 = data_types.BuildCrashStatsJobHistory()
    hour2.end_time_in_hours = 15
    hour2.put()

    hour1 = data_types.BuildCrashStatsJobHistory()
    hour1.end_time_in_hours = 10
    hour1.put()

    self.assertEqual(15, crash_stats.get_last_successful_hour())


def bq_convert_hour_to_index(hour, time_span, remainder):
  """Convert hour to index according to the SQL."""
  return (hour - remainder) // time_span


class IndexConversionTest(unittest.TestCase):
  """Test index conversion."""

  def _test(self, end_hour):
    """Test convert back and forth."""
    time_span = 24

    remainder = crash_stats.get_remainder_for_index(end_hour, time_span)
    end_index = bq_convert_hour_to_index(end_hour, time_span, remainder)
    self.assertEqual(
        end_hour,
        crash_stats.convert_index_to_hour(end_index, time_span, remainder))

    indices = []
    for h in range(end_hour - 24 * 5 + 1, end_hour + 1):
      indices.append(bq_convert_hour_to_index(h, time_span, remainder))

    hour_blocks = []
    for index in indices:
      hour_blocks.append(
          crash_stats.convert_index_to_hour(index, time_span, remainder))

    self.assertEqual(
        ([end_hour - 24 * 4] * 24 + [end_hour - 24 * 3] * 24 +
         [end_hour - 24 * 2] * 24 + [end_hour - 24] * 24 + [end_hour] * 24),
        hour_blocks)

  def test_convert_with_non_zero_remainder(self):
    """Test non-zero remainder."""
    self._test(24 * 10 + 13)
    self._test(24 * 10 + 1)
    self._test(24 * 10 - 1)

  def test_convert_with_zero_remainder(self):
    """Test zero remainder."""
    self._test(24 * 10)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Apps related functions."""

import os
import re
import time

from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

from . import adb
from . import constants

AAPT_CMD_TIMEOUT = 60
CHROME_CACHE_DIRS = [
    'app_chrome/*', 'app_tabs/*', 'app_textures/*', 'cache/*', 'files/*',
    'shared_prefs/*'
]
PACKAGES_THAT_CRASH_WITH_GESTURES = [
    'com.android.printspooler',
    'com.android.settings',
]
PACKAGE_OPTIMIZATION_INTERVAL = 30
PACKAGE_OPTIMIZATION_TIMEOUT = 30 * 60


def disable_packages_that_crash_with_gestures():
  """Disable known packages that crash on gesture fuzzing."""
  for package in PACKAGES_THAT_CRASH_WITH_GESTURES:
    adb.run_shell_command(['pm', 'disable-user', package], log_error=False)


def get_launch_command(app_args, testcase_path, testcase_file_url):
  """Get command to launch application with an optional testcase path."""
  application_launch_command = environment.get_value('APP_LAUNCH_COMMAND')
  if not application_launch_command:
    return ''

  package_name = get_package_name() or ''

  application_launch_command = application_launch_command.replace(
      '%APP_ARGS%', app_args)
  application_launch_command = application_launch_command.replace(
      '%DEVICE_TESTCASES_DIR%', constants.DEVICE_TESTCASES_DIR)
  application_launch_command = application_launch_command.replace(
      '%PKG_NAME%', package_name)
  application_launch_command = application_launch_command.replace(
      '%TESTCASE%', testcase_path)
  application_launch_command = application_launch_command.replace(
      '%TESTCASE_FILE_URL%', testcase_file_url)

  return application_launch_command


def get_package_name(apk_path=None):
  """Return package name."""
  # See if our environment is already set with this info.
  package_name = environment.get_value('PKG_NAME')
  if package_name:
    return package_name

  # See if we have the apk available to derive this info.
  if not apk_path:
    # Try getting apk path from APP_PATH.
    apk_path = environment.get_value('APP_PATH')
    if not apk_path:
      return None

  # Make sure that apk has the correct extension.
  if not apk_path.endswith('.apk'):
    return None

  # Try retrieving package name using aapt.
  aapt_binary_path = os.path.join(
      environment.get_platform_resources_directory(), 'aapt')
  aapt_command = '%s dump badging %s' % (aapt_binary_path, apk_path)
  output = adb.execute_command(aapt_command, timeout=AAPT_CMD_TIMEOUT)
  match = re.match('.*package: name=\'([^\']+)\'', output, re.DOTALL)
  if not match:
    return None
  return match.group(1)


def install(package_apk_path):
  """Install a package from an apk path."""
  return adb.run_command(['install', '-r', package_apk_path])


def is_installed(package_name):
  """Checks if the app is installed."""
  output = adb.run_shell_command(['pm', 'list', 'packages'])
  package_names = [line.split(':')[-1] for line in output.splitlines()]

  return package_name in package_names


def reset():
  """Reset to original clean state and kills pending instances."""
  package_name = get_package_name()
  if not package_name:
    return

  # Make sure package is actually installed.
  if not is_installed(package_name):
    return

  # Before clearing package state, save the minidumps.
  save_crash_minidumps(package_name)

  # Clean package state.
  adb.run_shell_command(['pm', 'clear', package_name])

  # Re-grant storage permissions.
  adb.run_shell_command(
      ['pm', 'grant', package_name, 'android.permission.READ_EXTERNAL_STORAGE'])
  adb.run_shell_command([
      'pm', 'grant', package_name, 'android.permission.WRITE_EXTERNAL_STORAGE'
  ])


def save_crash_minidumps(package_name):
  """Retain crash minidumps before app reset (chrome only)."""
  if package_name != 'com.google.android.apps.chrome':
    return

  # Ignore errors when running this command. Adding directory list check is
  # another adb call and since this is called frequently, we need to avoid that
  # extra call.
  adb.run_shell_command(
      ['cp', '/data/data/cache/Crash\\ Reports/*', constants.CRASH_DUMPS_DIR],
      log_error=False,
      root=True)


def stop():
  """Stop application and cleanup state."""
  package_name = get_package_name()
  if not package_name:
    return

  # Device can get silently restarted in case of OOM. So, we would need to
  # restart our shell as root in order to kill the application.
  adb.run_as_root()

  adb.kill_processes_and_children_matching_name(package_name)

  # Chrome specific cleanup.
  if package_name.endswith('.chrome'):
    cache_dirs_absolute_paths = [
        '/data/data/%s/%s' % (package_name, i) for i in CHROME_CACHE_DIRS
    ]
    save_crash_minidumps(package_name)
    adb.run_shell_command(
        ['rm', '-rf', ' '.join(cache_dirs_absolute_paths)], root=True)


def uninstall(package_name):
  """Uninstall a package given a name."""
  return adb.run_command(['uninstall', package_name])


def wait_until_optimization_complete():
  """Waits for package optimization to finish."""
  start_time = time.time()

  while time.time() - start_time < PACKAGE_OPTIMIZATION_TIMEOUT:
    package_optimization_finished = 'dex2oat' not in adb.get_ps_output()
    if package_optimization_finished:
      return

    logs.log('Waiting for package optimization to finish.')
    time.sleep(PACKAGE_OPTIMIZATION_INTERVAL)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Handler."""

import json
import os
import unittest

import flask
from flask import request
from flask import Response
import mock
import webtest
import yaml

from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from handlers import base_handler
from libs import auth
from libs import handler
from libs import helpers

_JSON_CONTENT_TYPE = 'application/json'


def mocked_db_config_get_value(key):
  """Return mocked values from db_config's get_value function."""
  if key == 'reproduce_tool_client_id':
    return 'ClientId'
  if key == 'reproduce_tool_client_secret':
    return 'Secret'
  return None


def mocked_load_yaml_file(yaml_file_path):
  """Return mocked version of local_config._load_yaml_file. Uses custom version
  of auth.yaml for tests in this file."""
  if os.path.basename(yaml_file_path) == 'auth.yaml':
    yaml_file_path = os.path.abspath(
        os.path.join(os.path.dirname(__file__), 'handler_data', 'auth.yaml'))

  return yaml.safe_load(open(yaml_file_path).read())


class JsonJsonPostHandler(base_handler.Handler):
  """Handler for json json post requests."""

  @handler.post(handler.JSON, handler.JSON)
  def post(self):
    test = request.get('test')
    response = Response()
    response.data = json.dumps({'data': test})
    response.status_code = 200
    return response


class FormHtmlPostHandler(base_handler.Handler):
  """Handler for form html post requests."""

  @handler.post(handler.FORM, handler.HTML)
  def post(self):
    test = request.form.get('test')
    response = Response()
    response.data = str(test)
    response.status_code = 200
    return response


class JsonGetHandler(base_handler.Handler):
  """Handler for json get requests."""

  @handler.get(handler.JSON)
  def get(self):
    test = request.get('test')
    response = Response()
    response.data = json.dumps({'data': test})
    response.status_code = 200
    return response


class HtmlGetHandler(base_handler.Handler):
  """Handler for html get requests."""

  @handler.get(handler.HTML)
  def get(self):
    test = request.get('test')
    response = Response()
    response.data = str(test)
    response.status_code = 200
    return response


class NeedsPrivilegeAccessHandler(base_handler.Handler):
  """Handler for needs privilege access requests."""

  @handler.get(handler.JSON)
  @handler.check_user_access(True)
  def get(self):
    return self.render_json({'data': 'with'})


class WithoutNeedsPrivilegeAccessHandler(base_handler.Handler):
  """Handler for without needs privilege access requests."""

  @handler.get(handler.JSON)
  @handler.check_user_access(False)
  def get(self):
    return self.render_json({'data': 'without'})


class CronHandler(base_handler.Handler):
  """Handler for crons."""

  @handler.cron()
  def get(self):
    return self.render_json({})


class CheckTestcaseAccessHandler(base_handler.Handler):
  """Handler for check testcase access requests."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.check_testcase_access
  def post(self, testcase):
    print("returning render_json")
    return self.render_json({'state': testcase.crash_state})


class CheckAdminAccessHandler(base_handler.Handler):
  """Handler for check admin access requests."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.check_admin_access
  def post(self):
    return self.render_json({'data': 'admin'})


class CheckAdminAccessIfOssFuzzHandler(base_handler.Handler):
  """Handler for check admin access if oss fuzz requests."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.check_admin_access_if_oss_fuzz
  def post(self):
    return self.render_json({})


class OAuthHandler(base_handler.Handler):
  """Handler for oauth requests."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.oauth
  def post(self):
    email = ''
    if auth.get_current_user():
      email = auth.get_current_user().email
    return self.render_json({'data': email})


class AllowedCorsHandler(base_handler.Handler):
  """Handler for allowed cors requests."""

  @handler.allowed_cors
  def post(self):
    return self.render_json({'data': 'yes'})


class CronTest(unittest.TestCase):
  """Test cron."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.config.db_config.get_value',
        'libs.form.generate_csrf_token',
        'libs.auth.is_current_user_admin',
        'libs.auth.get_current_user',
    ])
    self.mock.generate_csrf_token.return_value = 'csrf_token'
    self.mock.is_current_user_admin.return_value = False
    self.mock.get_current_user.return_value = auth.User('test@test.com')

    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=CronHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

  def test_succeed(self):
    """Test request from cron."""
    response = self.app.get('/', headers={'X-Appengine-Cron': 'True'})
    self.assertEqual(200, response.status_int)

  def test_fail(self):
    """Test request from non-cron."""
    response = self.app.get('/', expect_errors=True)
    self.assertEqual(403, response.status_int)


class PostTest(unittest.TestCase):
  """Test post wrapper"""

  def test_post_json_json(self):
    """Post JSON and receive JSON."""
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=JsonJsonPostHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.post_json('/', {'test': 123})
    self.assertEqual(_JSON_CONTENT_TYPE, resp.headers['Content-Type'])
    self.assertEqual(123, resp.json['data'])

  def test_post_json_json_failure(self):
    """Fail to post JSON."""
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=JsonJsonPostHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.post('/', {'test': 123}, expect_errors=True)
    self.assertEqual(_JSON_CONTENT_TYPE, resp.headers['Content-Type'])
    self.assertEqual(400, resp.status_int)

  def test_post_form_html(self):
    """Post Form-data and receive Html."""
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=FormHtmlPostHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.post('/', {'test': 123})
    self.assertNotEqual('application/json', resp.headers['Content-Type'])
    self.assertEqual(b'123', resp.body)


class GetTest(unittest.TestCase):
  """Test get wrapper."""

  def test_get_json(self):
    """Get and receive JSON."""
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=JsonGetHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.get('/', {'test': 123})
    self.assertEqual(_JSON_CONTENT_TYPE, resp.headers['Content-Type'])
    self.assertEqual('123', resp.json['data'])

  def test_get_html(self):
    """Get and receive Html."""
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=HtmlGetHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.get('/', {'test': 123})
    self.assertNotEqual('application/json', resp.headers['Content-Type'])
    self.assertEqual(b'123', resp.body)


class CheckUserAccessTest(unittest.TestCase):
  """Test check_user_access."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.access.has_access',
        'libs.helpers.get_user_email',
    ])

  def test_with_needs_privilege_access(self):
    """Test with needs_previlege_access."""
    self.mock.has_access.return_value = True
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/', view_func=NeedsPrivilegeAccessHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.get('/')
    self.assertEqual(200, resp.status_int)
    self.assertEqual('with', resp.json['data'])
    self.mock.has_access.assert_called_once_with(need_privileged_access=True)

  def test_without_needs_privilege(self):
    """Test without needs_previlege_access."""
    self.mock.has_access.return_value = True
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/', view_func=WithoutNeedsPrivilegeAccessHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.get('/')
    self.assertEqual(200, resp.status_int)
    self.assertEqual('without', resp.json['data'])
    self.mock.has_access.assert_called_once_with(need_privileged_access=False)

  def test_deny(self):
    """Test deny access."""
    self.mock.has_access.return_value = False
    self.mock.get_user_email.return_value = 'test@test.com'
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/', view_func=WithoutNeedsPrivilegeAccessHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.get('/', expect_errors=True)
    self.assertEqual(403, resp.status_int)
    self.assertEqual('', resp.json['message'])
    self.assertEqual('test@test.com', resp.json['email'])
    self.mock.has_access.assert_called_once_with(need_privileged_access=False)


class CheckTestcaseAccessTest(unittest.TestCase):
  """Test check_testcase_access."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.access.check_access_and_get_testcase',
    ])

  def test_no_testcase_id(self):
    """Test no testcase id."""
    self.mock.check_access_and_get_testcase.side_effect = (
        helpers.AccessDeniedException())
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/', view_func=CheckTestcaseAccessHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.post_json('/', {}, expect_errors=True)
    #print(resp)
    self.assertEqual(400, resp.status_int)
    self.assertRegex(resp.json['message'], '.*not a number.*')

  def test_invalid_testcase_id(self):
    """Test invalid testcase id."""
    self.mock.check_access_and_get_testcase.side_effect = (
        helpers.AccessDeniedException())
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/', view_func=CheckTestcaseAccessHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.post_json('/', {'testcaseId': 'aaa'}, expect_errors=True)
    self.assertEqual(400, resp.status_int)
    self.assertRegex(resp.json['message'], '.*not a number.*')

  def test_forbidden(self):
    """Test forbidden."""
    self.mock.check_access_and_get_testcase.side_effect = (
        helpers.AccessDeniedException())
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/', view_func=CheckTestcaseAccessHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.post_json('/', {'testcaseId': '123'}, expect_errors=True)
    self.assertEqual(403, resp.status_int)

  def test_allow(self):
    """Test allow."""
    testcase = data_types.Testcase()
    testcase.crash_state = 'state_value'
    self.mock.check_access_and_get_testcase.return_value = testcase
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/', view_func=CheckTestcaseAccessHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.post_json('/', {'testcaseId': '123'}, expect_errors=True)
    self.assertEqual(200, resp.status_int)
    self.assertEqual('state_value', resp.json['state'])


class CheckAdminAccessTest(unittest.TestCase):
  """Test check_testcase_access."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.auth.is_current_user_admin',
    ])

  def test_allowed(self):
    """Test allowing admin."""
    self.mock.is_current_user_admin.return_value = True
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=CheckAdminAccessHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.post_json('/', {})
    self.assertEqual(200, resp.status_int)
    self.assertEqual('admin', resp.json['data'])

  def test_forbidden(self):
    """Test allowing admin."""
    self.mock.is_current_user_admin.return_value = False
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=CheckAdminAccessHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.post_json('/', {}, expect_errors=True)
    self.assertEqual(403, resp.status_int)


class CheckAdminAccessIfOssFuzzTest(unittest.TestCase):
  """Test check_testcase_access_if_oss_fuzz."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.base.utils.is_oss_fuzz',
        'libs.auth.is_current_user_admin',
    ])
    test_helpers.patch_environ(self)
    self.mock.is_oss_fuzz.return_value = False

  def test_allowed_internal(self):
    """Test allowing non-admin and admin in internal."""
    self.mock.is_current_user_admin.return_value = False
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/', view_func=CheckAdminAccessIfOssFuzzHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.post_json('/', {})
    self.assertEqual(200, resp.status_int)

    self.mock.is_current_user_admin.return_value = True
    resp = self.app.post_json('/', {})
    self.assertEqual(200, resp.status_int)

  def test_allowed_oss_fuzz(self):
    """Test allowing admin in OSS-Fuzz."""
    self.mock.is_oss_fuzz.return_value = True
    self.mock.is_current_user_admin.return_value = True
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/', view_func=CheckAdminAccessIfOssFuzzHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.post_json('/', {})
    self.assertEqual(200, resp.status_int)

  def test_forbidden_oss_fuzz(self):
    """Test that non-admin in OSS-Fuzz are forbidden."""
    self.mock.is_oss_fuzz.return_value = True
    self.mock.is_current_user_admin.return_value = False
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/', view_func=CheckAdminAccessIfOssFuzzHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    resp = self.app.post_json('/', {}, expect_errors=True)
    self.assertEqual(403, resp.status_int)


class AllowOAuthTest(unittest.TestCase):
  """Test oauth."""

  def setUp(self):
    test_helpers.patch(self, ['libs.handler.get_email_and_access_token'])
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=OAuthHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)
    test_helpers.patch_environ(self)
    os.environ['AUTH_DOMAIN'] = 'localhost'

  def test_success(self):
    """Test setting environ and header properly."""
    self.mock.get_email_and_access_token.return_value = ('email', 'auth')

    resp = self.app.post_json(
        '/', {}, headers={'Authorization': 'Bearer AccessToken'})
    self.assertEqual(200, resp.status_int)
    self.assertEqual('email', resp.json['data'])
    self.assertEqual('auth',
                     resp.headers[handler.CLUSTERFUZZ_AUTHORIZATION_HEADER])
    self.assertEqual('email',
                     resp.headers[handler.CLUSTERFUZZ_AUTHORIZATION_IDENTITY])
    self.assertEqual(1, self.mock.get_email_and_access_token.call_count)
    self.mock.get_email_and_access_token.assert_has_calls(
        [mock.call('Bearer AccessToken')])

  def test_no_header(self):
    self.mock.get_email_and_access_token.return_value = ('email', 'auth')

    resp = self.app.post_json('/', {}, headers={})
    self.assertEqual(200, resp.status_int)
    self.assertEqual('', resp.json['data'])
    self.assertNotIn(handler.CLUSTERFUZZ_AUTHORIZATION_HEADER, resp.headers)
    self.assertEqual(0, self.mock.get_email_and_access_token.call_count)


class TestGetEmailAndAccessToken(unittest.TestCase):
  """Test get_email_and_access_token."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.config.db_config.get_value',
        'clusterfuzz._internal.config.local_config._load_yaml_file',
        'libs.handler.get_access_token',
        'requests.get',
    ])

    self.mock.get_value.side_effect = mocked_db_config_get_value
    self.mock._load_yaml_file.side_effect = mocked_load_yaml_file  # pylint: disable=protected-access

    config = local_config.AuthConfig()
    self.test_whitelisted_oauth_client_ids = config.get(
        'whitelisted_oauth_client_ids')
    self.test_whitelisted_oauth_emails = config.get('whitelisted_oauth_emails')

  def _assert_requests_get_call(self):
    self.assertEqual(1, self.mock.get.call_count)
    self.mock.get.assert_has_calls([
        mock.call(
            'https://www.googleapis.com/oauth2/v3/tokeninfo',
            params={'access_token': 'AccessToken'})
    ])
    self.mock.get.reset_mock()

  def test_allowed_bearer(self):
    """Test allowing Bearer."""
    for aud in self.test_whitelisted_oauth_client_ids:
      self.mock.get.return_value = mock.Mock(
          status_code=200,
          text=json.dumps({
              'aud': aud,
              'email': 'test@test.com',
              'email_verified': True
          }))

      email, token = handler.get_email_and_access_token('Bearer AccessToken')
      self.assertEqual('test@test.com', email)
      self.assertEqual('Bearer AccessToken', token)
      self._assert_requests_get_call()

  def test_allow_whitelised_accounts(self):
    """Test allow compute engine service account."""
    for email in self.test_whitelisted_oauth_emails:
      self.mock.get.reset_mock()
      self.mock.get.return_value = mock.Mock(
          status_code=200,
          text=json.dumps({
              'email_verified': True,
              'email': email
          }))

      returned_email, token = handler.get_email_and_access_token(
          'Bearer AccessToken')
      self.assertEqual(email, returned_email)
      self.assertEqual('Bearer AccessToken', token)
      self._assert_requests_get_call()

  def test_allowed_verification_code(self):
    """Test allowing VerificationCode."""
    self.mock.get.return_value = mock.Mock(
        status_code=200,
        text=json.dumps({
            'aud': 'ClientId',
            'email': 'test@test.com',
            'email_verified': True
        }))
    self.mock.get_access_token.return_value = 'AccessToken'

    email, token = handler.get_email_and_access_token('VerificationCode Verify')
    self.assertEqual('test@test.com', email)
    self.assertEqual('Bearer AccessToken', token)
    self.assertEqual(1, self.mock.get_access_token.call_count)
    self.mock.get_access_token.assert_has_calls([mock.call('Verify')])
    self._assert_requests_get_call()

  def test_invalid_authorization_header(self):
    """Test invalid authorization header."""
    with self.assertRaises(helpers.UnauthorizedException) as cm:
      handler.get_email_and_access_token('ReceiverAccessToken')

    self.assertEqual(401, cm.exception.status)
    self.assertEqual(
        'The Authorization header is invalid. It should have been started with'
        " 'Bearer '.", str(cm.exception))
    self.assertEqual(0, self.mock.get.call_count)

  def test_bad_status(self):
    """Test bad status."""
    self.mock.get.return_value = mock.Mock(status_code=403)

    with self.assertRaises(helpers.UnauthorizedException) as cm:
      handler.get_email_and_access_token('Bearer AccessToken')
    self.assertEqual(401, cm.exception.status)
    self.assertEqual(
        ('Failed to authorize. The Authorization header (Bearer AccessToken)'
         ' might be invalid.'), str(cm.exception))
    self._assert_requests_get_call()

  def test_invalid_json(self):
    """Test invalid json."""
    self.mock.get.return_value = mock.Mock(status_code=200, text='test')

    with self.assertRaises(helpers.EarlyExitException) as cm:
      handler.get_email_and_access_token('Bearer AccessToken')
    self.assertEqual(500, cm.exception.status)
    self.assertEqual('Parsing the JSON response body failed: test',
                     str(cm.exception))
    self._assert_requests_get_call()

  def test_invalid_client_id(self):
    """Test the invalid client id."""
    self.mock.get.return_value = mock.Mock(
        status_code=200,
        text=json.dumps({
            'aud': 'InvalidClientId',
            'email': 'test@test.com',
            'email_verified': False
        }))

    with self.assertRaises(helpers.EarlyExitException) as cm:
      handler.get_email_and_access_token('Bearer AccessToken')
    self.assertEqual(401, cm.exception.status)
    self.assertIn(
        "The access token doesn't belong to one of the allowed OAuth clients",
        str(cm.exception))
    self._assert_requests_get_call()

  def test_unverified_email(self):
    """Test unverified email."""
    self.mock.get.return_value = mock.Mock(
        status_code=200,
        text=json.dumps({
            'aud': 'ClientId',
            'email': 'test@test.com',
            'email_verified': False
        }))

    with self.assertRaises(helpers.EarlyExitException) as cm:
      handler.get_email_and_access_token('Bearer AccessToken')
    self.assertEqual(401, cm.exception.status)
    self.assertIn('The email (test@test.com) is not verified',
                  str(cm.exception))
    self._assert_requests_get_call()


class TestGetAccessToken(unittest.TestCase):
  """Test get_access_token."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.config.db_config.get_value',
        'clusterfuzz._internal.config.local_config._load_yaml_file',
        'requests.post',
    ])

    self.mock.get_value.side_effect = mocked_db_config_get_value
    self.mock._load_yaml_file.side_effect = mocked_load_yaml_file  # pylint: disable=protected-access

  def _assert_requests_post_call(self):
    self.assertEqual(1, self.mock.post.call_count)
    self.mock.post.assert_has_calls([
        mock.call(
            'https://www.googleapis.com/oauth2/v4/token',
            headers={'Content-Type': 'application/x-www-form-urlencoded'},
            data={
                'code': 'verify',
                'client_id': 'ClientId',
                'client_secret': 'Secret',
                'redirect_uri': 'urn:ietf:wg:oauth:2.0:oob',
                'grant_type': 'authorization_code'
            })
    ])

  def test_succeed(self):
    """Test succeed."""
    self.mock.post.return_value = mock.Mock(
        status_code=200, text=json.dumps({
            'access_token': 'token'
        }))

    token = handler.get_access_token('verify')
    self.assertEqual('token', token)
    self._assert_requests_post_call()

  def test_bad_status(self):
    """Test invalid_json."""
    self.mock.post.return_value = mock.Mock(status_code=403, text='test')

    with self.assertRaises(helpers.UnauthorizedException) as cm:
      handler.get_access_token('verify')
    self.assertEqual(401, cm.exception.status)
    self.assertEqual('Invalid verification code (verify): test',
                     str(cm.exception))
    self._assert_requests_post_call()

  def test_invalid_json(self):
    """Test invalid_json."""
    self.mock.post.return_value = mock.Mock(status_code=200, text='test')

    with self.assertRaises(helpers.EarlyExitException) as cm:
      handler.get_access_token('verify')
    self.assertEqual(500, cm.exception.status)
    self.assertEqual('Parsing the JSON response body failed: test',
                     str(cm.exception))
    self._assert_requests_post_call()


class AllowedCorsHandlerTest(unittest.TestCase):
  """Test allowed_cors."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.config.local_config._load_yaml_file',
    ])

    self.mock._load_yaml_file.side_effect = mocked_load_yaml_file  # pylint: disable=protected-access

    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=AllowedCorsHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

  def test_allow_cors(self):
    """Tests valid origins."""
    origins = [
        'http://test-client-site.appspot.com',
        'https://test-client-site-staging.appspot.com',
        'https://suborigin-dot-test-client-site.appspot.com',
        'http://suborigin-dot-test-client-site-staging.appspot.com',
    ]
    for origin in origins:
      resp = self.app.post_json('/', {}, headers={'Origin': origin})
      self.assertEqual(200, resp.status_int)
      self.assertEqual('yes', resp.json['data'])
      self.assertEqual(origin, resp.headers['Access-Control-Allow-Origin'])
      self.assertEqual('Origin', resp.headers['Vary'])
      self.assertEqual('true', resp.headers['Access-Control-Allow-Credentials'])
      self.assertEqual('GET,OPTIONS,POST',
                       resp.headers['Access-Control-Allow-Methods'])
      self.assertEqual('Accept,Authorization,Content-Type',
                       resp.headers['Access-Control-Allow-Headers'])
      self.assertEqual('3600', resp.headers['Access-Control-Max-Age'])

  def test_no_origin(self):
    """Tests no origin."""
    resp = self.app.post_json('/', {})
    self.assertEqual(200, resp.status_int)
    self.assertEqual('yes', resp.json['data'])
    self.assertIsNone(resp.headers.get('Access-Control-Allow-Origin'))

  def test_invalid_origin(self):
    """Tests no origin."""
    origins = [
        'http://bad-test-client-site.appspot.com',
        'https://bad-test-client-site-staging.appspot.com',
        'https://bad-test-client-site.appspot.com',
        'http://bad-test-client-site-staging.appspot.com',
    ]
    for origin in origins:
      resp = self.app.post_json('/', {'Origin': origin})
      self.assertEqual(200, resp.status_int)
      self.assertEqual('yes', resp.json['data'])
      self.assertIsNone(resp.headers.get('Access-Control-Allow-Origin'))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Log incoming reports of CSP violations."""

from flask import request

from clusterfuzz._internal.metrics import logs
from handlers import base_handler
from libs import handler
from libs import helpers


class ReportCspFailureHandler(base_handler.Handler):
  """Log failures on HTML pages caused by CSP."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.check_user_access(need_privileged_access=False)
  def post(self):
    """Handle a POST request."""
    report = request.get('csp-report')
    if not report:
      raise helpers.EarlyExitException('No CSP report.', 400)

    logs.log_error('CSP violation: {}'.format(report))
    return 'OK'
